from __future__ import print_function

import sys
import os
import shutil
import threading
import errno
import signal
import logging
import collections
import warnings
from datetime import datetime

if getattr(sys.version_info, 'major') == 2:
    import urllib2
elif getattr(sys.version_info, 'major') == 3:
    import urllib.request as urllib2
else:
    raise ImportError("Unable to detect the version of python you are using!")

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
	xrange = range

import numpy as np
import pandas as pd

from .extras import cprint
import gilly_utilities.datetime64
import gilly_utilities.manipulation

class DelayedKeyboardInterrupt(object):
    """Forces an operation to complete when this class function is called before an 
    operation. Handles Ctrl+C, Ctrl+Z and Ctrl+\ keyboard interrupts.
    
    Example
    -------
    
    with DelayedKeyboardInterrupt():
        for i in xrange(10):
            print(i)
            systime.sleep(1)
            
    >>> 0
    >>> 1
    >>> ^C[Warning] Please wait until the current operation has finish
    >>> ^Z[Warning] Please wait until the current operation has finish
    >>> ^\[Warning] Please wait until the current operation has finish
    >>> 2
    ...
    >>> 10
    >>> KeyboardInterrupt
    
    Reference
    ---------
    https://stackoverflow.com/a/21919644/8765762
    """    
    
    def __enter__(self):
        self.signal_received = False
        self.ctrlc_handler = signal.signal(signal.SIGINT, self.handler)
        self.ctrlz_handler = signal.signal(signal.SIGTSTP, self.handler)
        self.ctrlslash_handler = signal.signal(signal.SIGQUIT, self.handler)
        
    def handler(self, sig, frame):
        self.signal_received = (sig, frame)
        logging.debug('SIGINT received. Delaying KeyboardInterrupt.')
        cprint("[Warning] Please wait until the current operation has finished", type='warning')
        
    def __exit__(self, type, value, traceback):
        signal.signal(signal.SIGINT, self.ctrlc_handler)
        if self.signal_received:
            self.ctrlc_handler(*self.signal_received)
            
        signal.signal(signal.SIGTSTP, self.ctrlz_handler)
        if self.signal_received:
            self.ctrlz_handler(*self.signal_received)
            
        signal.signal(signal.SIGQUIT, self.ctrlslash_handler)
        if self.signal_received:
            self.ctrlslash_handler(*self.signal_received)

class unbuffered(object):
    """Forces the print statements to be written without the need for flushing. Useful for
    sending print statements to text files when running python on a cluster.
    
    Usage
    -----
    import sys
    import gilly_utilities as gu

    sys.stdout = gu.unbuffered(sys.stdout)
    
    Reference
    ---------
    https://stackoverflow.com/a/107717/8765762
    """
    
    def __init__(self, stream):
        self.stream = stream
    def write(self, data):
        self.stream.write(data)
        self.stream.flush()
    def writelines(self, datas):
        self.stream.writelines(datas)
        self.stream.flush()
    def __getattr__(self, attr):
        return getattr(self.stream, attr)

class antiverbose(object):
    """Suppresses the output of the sys.stdout.write and so no output
    will be visible on the console.
    
    To use this class place in code like so,
    
    with gu.antiverbose():
        ...
        
    ...
    
    References
    ----------
    https://stackoverflow.com/a/8391735/8765762
    https://stackoverflow.com/a/45669280/8765762
    """
    
    def __init__(self, verbose=False):
        """
        Specify whethere you want to turn on print statements or not
        """
        
        self.verbose = verbose
        
    def __enter__(self):
        
        if not self.verbose:
            self._original_stdout = sys.stdout
            sys.stdout = open(os.devnull, 'w')
            
    def __exit__(self, type, value, traceback):
    
        if not self.verbose:
            sys.stdout.close()
            sys.stdout = self._original_stdout

def verbose_class(func):
    """
    This decorator can be used to suppress the output of all stdout
    inside a class function. To toggle this function you need to set
    self.verbose as either True or False.
    
    Example
    -------
    >>> class fun(object):
    
    >>>     def __init__(self, verbose=False):
            self.verbose = verbose
            return
        
    >>> @dec_verbose_class
    >>> def func(self):
    >>>     print("Hello")
        
    >>> bob = fun(True)
    >>> bob.func()
    "Hello"
    >>> bob = fun(False)
    >>> bob.func()
    >>> 
    
    References
    ----------
    https://stackoverflow.com/a/5929165/8765762
    https://stackoverflow.com/a/7590709/8765762
    """
    
    def wrapper(self, *args, **kwargs):
        if not self.verbose:
        
            # block all printing to the console
            sys.stdout = open(os.devnull, 'w')
            # call the method in question
            value = func(self, *args, **kwargs)
            # enable all printing to the console
            sys.stdout = sys.__stdout__
            # pass the return value of the method back
            return value
        
        else:
            return func(self, *args, **kwargs)
    return wrapper

def isarray(arg):
    """Checks whether arg is array_like (e.g. tuple, list, np.ndarray, etc.)"""
    
    #return type(arg) is not str and isinstance(arg, collections.Sequence)
    
    dtypes = [list, tuple, np.ndarray, pd.DataFrame]
    return np.any([isinstance(arg, dtype) for dtype in dtypes])

def isnumeric(arg, warn=True):
    """
    Test element-wise for numerics (e.g. 2.4, 2).

    The result is returned as a boolean array.
    """
    
    # Force arg to be a numpy array
    arg = np.atleast_1d(arg)
    
    # Check for finitness
    try:
        return np.isfinite(arg)
    except TypeError:
        if warn is True:
            warnings.warn("[gu.isnumeric] invalid data type found in arg. Returning False", UserWarning, stacklevel=2)
        return np.array(False)

def isnone(arg, keepshape=True):
    """
    Test element-wise for nones.

    The result is returned as a boolean array.
    """

    # Force arg to be a numpy array
    arg = np.atleast_1d(arg)
    nones_full = np.zeros(arg.shape, dtype=bool)
    
    # Create index array of arg to reorganise after we need to flatten arg
    index = np.arange(arg.size).reshape(arg.shape)
    
    # Flatten arg array
    arg = np.hstack(arg.flat)
    index_flat = np.hstack(index)
    
    # Check for nones
    nones = np.array([val is None for val in arg], dtype=bool)
    
    if keepshape is True:
        
        # Return a restructured array
        for ind, val in zip(index_flat, nones):
            nones_full[np.where(index == ind)] = val
        
        return nones_full
    
    else:
        return nones

def iszero(arg, keepshape=True):
    """
    Test element-wise for zeros.
    
    The result is returned as a boolean array.
    """
    
    # Force arg to be a numpy array
    arg = np.atleast_1d(arg)
    zero_full = np.zeros(arg.shape, dtype=bool)
    
    # Create index array of arg to reorganise after we need to flatten arg
    index = np.arange(arg.size).reshape(arg.shape)
    
    # Flatten arg array
    arg = np.hstack(arg.flat)
    index_flat = np.hstack(index)
    
    # Check for zeros
    zeros = np.array([val == 0 for val in arg], dtype=bool)
    
    if keepshape is True:
        
        # Return a restructured array
        for ind, val in zip(index_flat, zeros):
            zero_full[np.where(index == ind)] = val
        
        return zero_full
    
    else:
        return zeros
        
def isval(arg, value, keepshape=True):
    """
    Test element-wise for specific value.
    
    The result is returned as a boolean array.
    """
    
    # Force arg to be a numpy array
    arg = np.atleast_1d(arg)
    zero_full = np.zeros(arg.shape, dtype=bool)
    
    # Check if arg has any data
    if arg.size == 0:
        return False
    
    # Create index array of arg to reorganise after we need to flatten arg
    index = np.arange(arg.size).reshape(arg.shape)
    
    # Flatten arg array
    arg = np.hstack(arg.flat)
    index_flat = np.hstack(index)
    
    # Check for zeros
    zeros = np.array([val == value for val in arg], dtype=bool)
    
    if keepshape is True:
        
        # Return a restructured array
        for ind, val in zip(index_flat, zeros):
            zero_full[np.where(index == ind)] = val
        
        return zero_full
    
    else:
        return zeros
    
"""
Works out the length of string arrays element-wise

Reference
---------
https://stackoverflow.com/a/44587916/8765762
"""
arrlen = np.vectorize(len)
    
def ensure_dir(dirname, dir_or_file='dir'):
    """Ensure that a named directory exists; if it does not, attempt to create it.
    
    Sources
    -------
    http://stackoverflow.com/a/21349806
    """

    # Copy dirname to return to user
    dirname_return = dirname

    try:
        os.makedirs(dirname) if dir_or_file == 'dir' else os.makedirs(os.path.dirname(dirname))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    return dirname

def del_folder(folder, sub=False): 
    """Deletes a folder. Option sub to make recursive"""

    for the_file in os.listdir(folder):
        file_path = os.path.join(folder, the_file)
        try:
            if os.path.isfile(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path) and sub == True: 
                shutil.rmtree(file_path)
        except Exception as e:
            print(e)
            
def Print_Output(delay_time, t_start, it_tot, it):
  t = threading.Timer(delay_time, Print_Output(delay_time, t_start, it_tot, it)).start()
  os.system('cls' if os.name=='nt' else 'clear')
  print("Time Left (s): ", "%.0f" % ((it_tot-it)*((time.time()-t_start)/it)))
  print("Events Left: ", it_tot-it)

def file_checker(dates, date_range, files=None, file_size_min=0, yes_file=1,
        no_file=0, corrupt_file=2):
    """Checks the availability of files
    
    Parameters
    ----------
    dates : numpy object array
        An array of datetimes corresponding to each file location in files.
    date_range : nump object array
        An array over the list of dates you want to test.
    files : numpy array
        A list of file locations to be tested
    file_size_min : int, optional
        The minimum file size (in bytes) required to be valid. This is useful for files that
        exist but don't contain any data.

    Returns
    -------
    File_Availability : list
        A boolean list with 1 for file exists and is above the minimum file size threshold
        and zero for either file doesn't exist or file is below or equal to minimum file
        threshold.        
    """
    
    # if files are not specified, then only check dates occur in dates range.
    # This method is much quicker, as you can use numpy element-wise
    # operations and don't have to check file size of every file.
    if files is None:
        # Create zero array matching date_range length
        File_Availability = np.zeros(date_range.size)
        
        # If date is found in date_range then set File_Availability to 1
        File_Availability[np.in1d(date_range, dates)] = 1
        
        # Return data
        return File_Availability
    else:
        # Create zero array matching date_range length
        File_Availability = np.zeros(date_range.size)
        File_Availability = []
        for i in xrange(date_range.shape[0]):
            if np.sum(dates == date_range[i]) == 1:            
                if os.stat(files[dates == date_range[i]][0]).st_size > file_size_min:
                    File_Availability.append(yes_file)
                else:
                    File_Availability.append(corrupt_file)
            else:
                File_Availability.append(no_file)
                
        return np.array(File_Availability, dtype=float)

def file_checker2(dates, date_range, files=None, file_size_min=0, yes_file=1,
        no_file=0, corrupt_file=2):
    """Checks the availability of files
    
    Parameters
    ----------
    dates : numpy object array
        An array of datetimes corresponding to each file location in files.
    date_range : nump object array
        An array over the list of dates you want to test.
    files : numpy array
        A list of file locations to be tested
    file_size_min : int, optional
        The minimum file size (in bytes) required to be valid. This is useful for files that
        exist but don't contain any data.

    Returns
    -------
    File_Availability : list
        A boolean list with 1 for file exists and is above the minimum file size threshold
        and zero for either file doesn't exist or file is below or equal to minimum file
        threshold.        
    """

    # Create zero array matching date_range length
    File_Availability = np.zeros(date_range.size)
    
    # Get mask for dates in date_range
    mask_date = np.in1d(date_range, dates)

    # If date is found in date_range then set File_Availability to yes_file
    File_Availability[mask_date] = yes_file
        
    # if files are not specified, then only check dates occur in dates range.
    # This method is much quicker, as you can use numpy element-wise
    # operations and don't have to check file size of every file.
    if files is not None:

        # Get file size for all files
        file_sizes = []
        for file in files:
            try:
                file_sizes.append(os.stat(file).st_size)
            except OSError:
                file_sizes.append(0)
        
        # Convert to numpy array to use element-wise ops.
        file_sizes = np.array(file_sizes)
        
        # Test which files are smaller than expected
        mask_filesize = file_sizes <= file_size_min

        # Get date_range for dates that have files too small
        mask_datesfile = np.in1d(date_range, dates[mask_filesize])

        # Combine mask_filesize with mask_date
        File_Availability[mask_datesfile] = corrupt_file
    
    # Return data
    return File_Availability
    
def File_Aligner_NoSize(files, dates, date_range):
    """Checks the availability of files
    
    Parameters
    ----------
    files : numpy array
        A list of file locations to be tested
    dates : numpy object array
        An array of datetimes corresponding to each file location in files.
    date_range : nump object array
        An array over the list of dates you want to test.
    file_size_min : int, optional
        The minimum file size (in bytes) required to be valid. This is useful for files that
        exist but don't contain any data.

    Returns
    -------
    File_Availability : list
        A boolean list with 1 for file exists and is above the minimum file size threshold
        and zero for either file doesn't exist or file is below or equal to minimum file
        threshold.        
    """
    
    File_Date = np.zeros(date_range.shape[0], dtype=object)
    File_Loc = np.zeros(date_range.shape[0], dtype='S256')
    for i in xrange(date_range.shape[0]):
        if np.sum(dates == date_range[i]) >= 1:            
            File_Date[i] = date_range[i]
            File_Loc[i] = files[dates == date_range[i]][0]
            
    return np.array(File_Loc, dtype='S256'), np.array(File_Date, dtype=object)
    
def File_Aligner(files, dates, date_range, file_size_min=0):
    """Checks the availability of files
    
    Parameters
    ----------
    files : numpy array
        A list of file locations to be tested
    dates : numpy object array
        An array of datetimes corresponding to each file location in files.
    date_range : nump object array
        An array over the list of dates you want to test.
    file_size_min : int, optional
        The minimum file size (in bytes) required to be valid. This is useful for files that
        exist but don't contain any data.

    Returns
    -------
    File_Availability : list
        A boolean list with 1 for file exists and is above the minimum file size threshold
        and zero for either file doesn't exist or file is below or equal to minimum file
        threshold.        
    """
    
    File_Date = np.zeros(date_range.shape[0], dtype=object)
    File_Loc = np.zeros(date_range.shape[0], dtype='S256')
    for i in xrange(date_range.shape[0]):
        if np.sum(dates == date_range[i]) == 1:    
            if os.stat(files[dates == date_range[i]][0]).st_size > file_size_min:
                File_Date[i] = date_range[i]
                File_Loc[i] = files[dates == date_range[i]][0]
            else:
                File_Date[i] = 2
                File_Loc[i] = 2
            
    return np.array(File_Loc, dtype='S256'), np.array(File_Date, dtype=object)
    
def urllib_authentication(url, username, password, library='urllib2'):
    """
    When using urllib2 on websites which require authentication. This
    function can be used to access the website
    """
    
    p = urllib2.HTTPPasswordMgrWithDefaultRealm()

    p.add_password(None, url, username, password)

    handler = urllib2.HTTPBasicAuthHandler(p)
    opener = urllib2.build_opener(handler)
    urllib2.install_opener(opener)
    
    return


import numpy as np
import scipy as sp
import statsmodels.api as sm
import sys
import warnings
try:
    from sklearn.linear_model import TheilSenRegressor, HuberRegressor
    Huber = True
except ImportError:
    from sklearn.linear_model import TheilSenRegressor
    Huber = False

from collections import namedtuple    

from .system import isarray
from .manipulation import stride, broadcast, digitize, searchsorted
from .nanfunctions import antinan, antifinite
from .math import power

# Import statistical bootstrapping library
from scikits.bootstrap import bootstrap as boot

femto = sys.float_info.epsilon

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
	xrange = range
	
__version__ = 1.1

HuberRegressionResult = namedtuple('HuberRegressionResult',
                                  ('slope', 'intercept', 'rvalue',
                                   'pvalue', 'stderr'))

DescriptiveStats_Extend = namedtuple('DescriptiveStats',
                                    ('count', 'min', 'max', 'mean', 'median',
                                     'std', 'sum', 'range', 'noise',
                                     'percentile_5th', 'percentile_10th',
                                     'percentile_25th', 'percentile_75th',
                                     'percentile_90th', 'percentile_95th'))
                                                    
DescriptiveStats = namedtuple('DescriptiveStats', 
                             ('count', 'min', 'max', 'mean', 'median', 'std'))

### Covariance Methods ###
#Below is a list of covariance algorithms which have been found from
#wikipedia (https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Covariance)
#and from other sources.

def naive_covariance(data1, data2):
    """Fast but prone to catastrophic cancellaion when
    E[XY] ~ E[X]E[Y] i.e. the numbers are small and
    not negligible to the computer precision limit"""

    n = len(data1)
    sum12 = 0
    sum1 = sum(data1)
    sum2 = sum(data2)

    for i1, i2 in zip(data1, data2):
        sum12 += i1*i2

    covariance = (sum12 - sum1*sum2 / n) / (n - 1)
    
    return covariance
    
def shifted_data_covariance(dataX, dataY):
    """Fixes catastrophic cancellation probelem"""

    n = len(dataX)
    if (n < 2):
        return 0
    Kx = dataX[0]
    Ky = dataY[0]
    Ex = Ey = Exy = 0
    for iX, iY in zip(dataX, dataY):
        Ex += iX - Kx
        Ey += iY - Ky
        Exy += (iX - Kx) * (iY - Ky)
    
    return (Exy - Ex * Ey / n) / n
    
def shifted_data_covariance_numpy(dataX, dataY, elem=None, fill_edges=False):
    """Expects numpy arrays input but will calculate the
    covariance much quicker
    
    N.B. This is using the biased estimator version where
    we normalised the variance by the number of values in
    our array (i.e. n). For an unbiased estimator change 
    the last n in the return line from n to n-1
    
    Parameters
    ----------
    elem : int, optional
        elem is not None then dataX and dataY is subset in
        arrays of length elem and the covariance of each
        subset is found.
    fill_edges : bool, optional
        This module will split the data and centre on the
        parent dataset. Therefore, this option will buffer
        the covariance array with 0's to keep the dimensions
        the same
    """
    
    n = dataX.size
    if n < 2: return np.nan
    
    if elem is None:
        
        Ex     = sum(dataX - dataX[0])
        Ey     = sum(dataY - dataY[0])
        Exy = sum((dataX - dataX[0])*(dataY - dataY[0]))
        
        return (Exy - Ex * Ey / n) / n
    
    else:
        
        as_strided = np.lib.stride_tricks.as_strided #Numpy array broadcasting trick
        strides = dataX.strides[0]                     #Used for memory checking of array 
        ncols = (n - elem)
        nrows = elem
        
        #Split dataX in array of elem elements and determine the mean subset offset by the first element of each subset
        #dataXX = np.array(zip(*(dataX[i:] for i in xrange(elem)))).T
        #dataXX = np.array([dataX[i:i+elem] for i in xrange(dataX.size-elem+1)]).T
        dataXX = as_strided(dataX, (ncols, nrows), (strides, strides)).copy().T
        Ex = sum(dataXX - dataX[:-elem])
        
        #Split dataY in array of elem elements and determine the mean subset offset by the first element of each subset
        #dataYY = np.array(zip(*(dataY[i:] for i in xrange(elem)))).T
        #dataYY = np.array([dataY[i:i+elem] for i in xrange(dataY.size-elem+1)]).T
        dataYY = as_strided(dataY, (ncols, nrows), (strides, strides)).copy().T
        Ey = sum(dataYY - dataY[:-elem])
        
        #Calculate the sum of means for each subset
        Exy = sum((dataXX - dataX[:-elem])*(dataYY - dataY[:-elem]))
        
        cov = (Exy - Ex * Ey / elem) / elem
        
        if fill_edges is True:
            return list_func(cov).rljust(dataX.size, 0, float)
        else:
            return cov
           
def two_pass_covariance(data1, data2):
    """Two pass algorithm computes the sample means
    and then the covariance"""

    n = len(data1)

    mean1 = sum(data1) / n
    mean2 = sum(data2) / n

    covariance = 0

    for i1, i2 in zip(data1, data2):
        a = i1 - mean1
        b = i2 - mean2
        covariance += a*b / n
    
    return covariance
    
def Lineariser_Power(endog, exog, weights=None, initial=1, maxits=1000, prec=10**-7):
    """Linearises exog (x axis) using a power law setup and iterates
    the power function b. The newly manipulated data is then solved
    using an OLS where the remaining a and c parameters can be solved
    following the format:
    
                        y = a*x^b + c

    NOTE 1 30/08/17: NumPy is unable to raise float by a fractional float
    (e.g. -0.455**0.9) which will return a np.nan. To overcome this we've
    had to temporaily convert the array exog to a complex function and
    then solve and combine the real and imaginary parts using the np.
    absolute function. This provides the correct solution. On the down
    side this will have ramifications for computational speeds. No 
    solution really exists for this issue. 
    
    see for info: https://stackoverflow.com/a/17747293
    
    """
    
    iter = initial
    sw = 1
    iter_all = []
    sw_all = []
    for i in xrange(maxits):
        
        X2 = np.column_stack((np.ones(len(exog)), complex2float(np.array(exog, dtype=complex)**(iter-0.1/sw))))
        model = sm.OLS(endog, X2) if weights is None else sm.WLS(endog, X2, weights=weights) 
        res_min = model.fit().rsquared**0.5
        
        X2 = np.column_stack((np.ones(len(exog)), complex2float(np.array(exog, dtype=complex)**iter)))
        model = sm.OLS(endog, X2) if weights is None else sm.WLS(endog, X2, weights=weights) 
        res = model.fit().rsquared**0.5
        
        X2 = np.column_stack((np.ones(len(exog)), complex2float(np.array(exog, dtype=complex)**(iter+0.1/sw))))
        model = sm.OLS(endog, X2) if weights is None else sm.WLS(endog, X2, weights=weights) 
        res_plus = model.fit().rsquared**0.5
        
        if (res_min > res) &  (res > res_plus):
            iter -= 0.1/sw
            if sw <= 1: sw /= 2.
        elif (res_min < res) &  (res < res_plus):
            iter += 0.1/sw
            if sw <= 1: sw /= 2.
        elif (res_min > res) &  (res < res_plus):
            if (res_min - res) > (res_plus - res):
                iter -= 0.01/sw
            else:
                iter += 0.01/sw
            sw += 1
        elif (res_min < res) &  (res > res_plus):
            if (res - res_min) > (res - res_plus):
                iter += 0.01/sw
            else:
                iter -= 0.01/sw
            sw  += 1
        iter_all.append(iter)
        sw_all.append(sw)
        
        if np.abs((res - res_min) + (res - res_plus)) < prec: break
    if i == maxits-1: warnings.warn("Maximum iterations (%.0f) reached and precision of the R squared value was not found!" % (maxits), UserWarning, stacklevel=2)
    
    X2 = np.column_stack((complex2float(np.array(exog, dtype=complex)**iter), np.ones(len(exog))))
    model = sm.OLS(endog, X2) if weights is None else sm.WLS(endog, X2, weights=weights) 
    res = model.fit()
    
    return res, iter, X2, endog
    
def Lineariser_Power_v2(endog, exog, initial=1, maxits=1000, prec=10**-7, weights=None, std_err=False):
    """Linearises exog (x axis) using a power law set-up and iterates
    the power function b. The newly manipulated data is then solved
    using an OLS where the remaining a and c parameters can be solved
    following the format:
    
                        y = a*x^b + c

    NOTE 1 30/08/17: NumPy is unable to raise float by a fractional float
    (e.g. -0.455**0.9) which will return a np.nan. To overcome this we've
    had to temporarily convert the array exog to a complex function and
    then solve and combine the real and imaginary parts using the np.
    absolute function. This provides the correct solution. On the down
    side this will have ramifications for computational speeds. No 
    solution really exists for this issue. 
    
    see for info: https://stackoverflow.com/a/17747293
    
    Now added in error analysis for the power law parameter. This is
    achieved using bootstrapping, similar to EPCC_PC_Estimator. The one
    large downside is the time it takes to compute several 100 bootstraps
    which are typically required for an appropriate estimator or the power
    standard error. 
    
    std_err : int, optional
        the number of bootstrap iterations required for power error 
        analysis.    
    """
    
    Power = initial
    sw = 1
    Power_All = []
    sw_all = []
    for i in xrange(maxits):
        
        X2 = np.column_stack((complex2float(np.array(exog, dtype=complex)**(Power-0.1/sw)), np.ones(len(exog))))
        #print("np.column_stack((endog, X2, weights))", np.column_stack((endog, X2, weights)).shape)
        #sys.exit()
        YX = antifinite(np.column_stack((endog, X2)).T)    if weights is None else antifinite(np.column_stack((endog, X2, weights)).T)
        model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
        res_min = model.fit().rsquared
        
        X2 = np.column_stack((complex2float(np.array(exog, dtype=complex)**Power), np.ones(len(exog))))
        YX = antifinite(np.column_stack((endog, X2)).T)    if weights is None else antifinite(np.column_stack((endog, X2, weights)).T)        
        model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
        res = model.fit().rsquared
        
        X2 = np.column_stack((complex2float(np.array(exog, dtype=complex)**(Power+0.1/sw)), np.ones(len(exog))))
        YX = antifinite(np.column_stack((endog, X2)).T)    if weights is None else antifinite(np.column_stack((endog, X2, weights)).T)        
        model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1])  
        res_plus = model.fit().rsquared
        
        if (res_min > res) &  (res > res_plus):
            Power -= 0.1/sw
            if sw <= 1: sw /= 2.
        elif (res_min < res) &  (res < res_plus):
            Power += 0.1/sw
            if sw <= 1: sw /= 2.
        elif (res_min > res) &  (res < res_plus):
            if (res_min - res) > (res_plus - res):
                Power -= 0.01/sw
            else:
                Power += 0.01/sw
            sw += 1
        elif (res_min < res) &  (res > res_plus):
            if (res - res_min) > (res - res_plus):
                Power += 0.01/sw
            else:
                Power -= 0.01/sw
            sw  += 1
        Power_All.append(Power)
        sw_all.append(sw)
        
        if np.abs((res - res_min) + (res - res_plus)) < prec: break
    if i == maxits-1: warnings.warn("Maximum iterations (%.0f) reached and precision of the R squared value was not found!" % (maxits), UserWarning, stacklevel=2)
    
    X2 = np.column_stack((complex2float(np.array(exog, dtype=complex)**Power), np.ones(len(exog))))
    YX = antifinite(np.column_stack((endog, X2)).T)    if weights is None else antifinite(np.column_stack((endog, X2, weights)).T)
    model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
    res = model.fit()
    Power_Best = Power
    
    #Perform error analysis using bootstrap
    if std_err is not False: 
        Residuals = endog - res.fittedvalues
        Results_All = np.zeros(std_err, dtype=object)
        
        for jj in xrange(std_err):
            Bootstrap = np.array([value + Residuals[np.random.randint(exog.size)] for value in endog], dtype=float)
            
            #Rerun power lineariser
            Power = initial
            sw = 1
            Power_All = []
            sw_all = []
            for i in xrange(maxits):
                
                X2 = np.column_stack((complex2float(np.array(exog, dtype=complex)**(Power-0.1/sw)), np.ones(len(exog))))
                YX = antifinite(np.column_stack((Bootstrap, X2)).T)    if weights is None else antifinite(np.column_stack((endog, X2, weights)).T)        
                model = sm.OLS(YX[:,0], YX[:,1:])if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
                res_min = model.fit().rsquared
                
                X2 = np.column_stack((complex2float(np.array(exog, dtype=complex)**Power), np.ones(len(exog))))
                YX = antifinite(np.column_stack((Bootstrap, X2)).T)    if weights is None else antifinite(np.column_stack((endog, X2, weights)).T)        
                model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
                res = model.fit().rsquared
                
                X2 = np.column_stack((complex2float(np.array(exog, dtype=complex)**(Power+0.1/sw)), np.ones(len(exog))))
                YX = antifinite(np.column_stack((Bootstrap, X2)).T)    if weights is None else antifinite(np.column_stack((endog, X2, weights)).T)        
                model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
                res_plus = model.fit().rsquared
                
                if (res_min > res) &  (res > res_plus):
                    Power -= 0.1/sw
                    if sw <= 1: sw /= 2.
                elif (res_min < res) &  (res < res_plus):
                    Power += 0.1/sw
                    if sw <= 1: sw /= 2.
                elif (res_min > res) &  (res < res_plus):
                    if (res_min - res) > (res_plus - res):
                        Power -= 0.01/sw
                    else:
                        Power += 0.01/sw
                    sw += 1
                elif (res_min < res) &  (res > res_plus):
                    if (res - res_min) > (res - res_plus):
                        Power += 0.01/sw
                    else:
                        Power -= 0.01/sw
                    sw  += 1
                Power_All.append(Power)
                sw_all.append(sw)
                
                if np.abs((res - res_min) + (res - res_plus)) < prec: break
            if i == maxits-1: warnings.warn("Maximum iterations (%.0f) reached and precision of the R squared value was not found!" % (maxits), UserWarning, stacklevel=2)
            
            Results_All[jj] = Power
        
        #Determine the 95% confidence limits
        Power_Error = np.diff(np.percentile(Results_All, [5,95])).astype(float)/2
        
        return res, [Power_Best, Power_Error], YX[:,1:], YX[:,0]
        
    return res, Power_Best, YX[:,1:], YX[:,0]

def Lineariser_Power_v3(endog, exog, initial=1, maxits=1000, prec=10**-7, weights=None, std_err=False):
    """Linearises exog (x axis) using a power law set-up and iterates
    the power function b. The newly manipulated data is then solved
    using an OLS where the remaining a and c parameters can be solved
    following the format:
    
                        y = a*x^b + c

    NOTE 1 30/08/17: NumPy is unable to raise float by a fractional float
    (e.g. -0.455**0.9) which will return a np.nan. To overcome this we've
    had to temporarily convert the array exog to a complex function and
    then solve and combine the real and imaginary parts using the np.
    absolute function. This provides the correct solution. On the down
    side this will have ramifications for computational speeds. No 
    solution really exists for this issue. 
    
    see for info: https://stackoverflow.com/a/17747293
    
    Now added in error analysis for the power law parameter. This is
    achieved using bootstrapping, similar to EPCC_PC_Estimator. The one
    large downside is the time it takes to compute several 100 bootstraps
    which are typically required for an appropriate estimator or the power
    standard error. 
    
    std_err : int, optional
        the number of bootstrap iterations required for power error 
        analysis.    
    """
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        
        #Inital Conditions
        coeff_power = initial
        sw = 1
        Power_All = []
        sw_all = []
        
        #Optimise power coefficient
        for i in xrange(maxits):
            X2 = np.column_stack((power(exog, coeff_power-0.1/sw), np.ones(len(exog))))
            YX = antifinite(np.column_stack((endog, X2)), oldmethod=False) if weights is None else antifinite(np.column_stack((endog, X2, weights)), oldmethod=False)

            model1 = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
            res_min = model1.fit().rsquared
            
            X2 = np.column_stack((power(exog, coeff_power), np.ones(len(exog))))
            YX = antifinite(np.column_stack((endog, X2)), oldmethod=False) if weights is None else antifinite(np.column_stack((endog, X2, weights)), oldmethod=False)        
            model2 = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
            res = model2.fit().rsquared
            
            X2 = np.column_stack((power(exog, coeff_power+0.1/sw), np.ones(len(exog))))
            YX = antifinite(np.column_stack((endog, X2)), oldmethod=False) if weights is None else antifinite(np.column_stack((endog, X2, weights)), oldmethod=False)        
            model3 = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1])  
            res_plus = model3.fit().rsquared
            
            if (res_min > res) &  (res > res_plus):
                coeff_power -= 0.1/sw
                if sw <= 1: sw /= 2.
            elif (res_min < res) &  (res < res_plus):
                coeff_power += 0.1/sw
                if sw <= 1: sw /= 2.
            elif (res_min > res) &  (res < res_plus):
                if (res_min - res) > (res_plus - res):
                    coeff_power -= 0.01/sw
                else:
                    coeff_power += 0.01/sw
                sw += 1
            elif (res_min < res) &  (res > res_plus):
                if (res - res_min) > (res - res_plus):
                    coeff_power += 0.01/sw
                else:
                    coeff_power -= 0.01/sw
                sw  += 1
            Power_All.append(coeff_power)
            sw_all.append(sw)
            
            if np.abs((res - res_min) + (res - res_plus)) < prec: break
            
        if i == maxits-1: warnings.warn("Maximum iterations (%.0f) reached and precision of the R squared value was not found!" % (maxits), UserWarning, stacklevel=2)
        
        X2 = np.column_stack((power(exog, coeff_power), np.ones(len(exog))))
        YX = antifinite(np.column_stack((endog, X2)), oldmethod=False)    if weights is None else antifinite(np.column_stack((endog, X2, weights)), oldmethod=False)

        model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
        res_final = model.fit()
        Power_Best = coeff_power
        
        #Perform error analysis using bootstrap
        if std_err is not False: 
            Residuals = endog - res_final.fittedvalues
            Results_All = np.zeros(std_err, dtype=object)
            
            for jj in xrange(std_err):
                Bootstrap = np.array([value + Residuals[np.random.randint(exog.shape[0])] for value in endog], dtype=float)
                
                #Rerun power lineariser
                coeff_power = initial
                sw = 1
                Power_All = []
                sw_all = []
                for i in xrange(maxits):
                    
                    X2 = np.column_stack((power(exog, coeff_power-0.1/sw), np.ones(len(exog))))
                    YX = antifinite(np.column_stack((Bootstrap, X2)), oldmethod=False)    if weights is None else antifinite(np.column_stack((endog, X2, weights)), oldmethod=False)        
                    model = sm.OLS(YX[:,0], YX[:,1:])if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
                    res_min = model.fit().rsquared
                    
                    X2 = np.column_stack((power(exog, coeff_power), np.ones(len(exog))))
                    YX = antifinite(np.column_stack((Bootstrap, X2)), oldmethod=False)    if weights is None else antifinite(np.column_stack((endog, X2, weights)), oldmethod=False)        
                    model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
                    res = model.fit().rsquared
                    
                    X2 = np.column_stack((power(exog, coeff_power+0.1/sw), np.ones(len(exog))))
                    YX = antifinite(np.column_stack((Bootstrap, X2)), oldmethod=False)    if weights is None else antifinite(np.column_stack((endog, X2, weights)), oldmethod=False)        
                    model = sm.OLS(YX[:,0], YX[:,1:]) if weights is None else sm.WLS(YX[:,0], YX[:,1:-1], weights=YX[:,-1]) 
                    res_plus = model.fit().rsquared
                    
                    if (res_min > res) &  (res > res_plus):
                        coeff_power -= 0.1/sw
                        if sw <= 1: sw /= 2.
                    elif (res_min < res) &  (res < res_plus):
                        coeff_power += 0.1/sw
                        if sw <= 1: sw /= 2.
                    elif (res_min > res) &  (res < res_plus):
                        if (res_min - res) > (res_plus - res):
                            coeff_power -= 0.01/sw
                        else:
                            coeff_power += 0.01/sw
                        sw += 1
                    elif (res_min < res) &  (res > res_plus):
                        if (res - res_min) > (res - res_plus):
                            coeff_power += 0.01/sw
                        else:
                            coeff_power -= 0.01/sw
                        sw  += 1
                    Power_All.append(coeff_power)
                    sw_all.append(sw)
                    
                    if np.abs((res - res_min) + (res - res_plus)) < prec: break
                if i == maxits-1: warnings.warn("Maximum iterations (%.0f) reached and precision of the R squared value was not found!" % (maxits), UserWarning, stacklevel=2)
                
                Results_All[jj] = coeff_power
            
            #Determine the 95% confidence limits
            Power_Error = np.diff(np.percentile(Results_All, [2.5,97.5])).astype(float)/2
            
            return res_final, [Power_Best, Power_Error], YX[:,1:], YX[:,0]
        
    return res_final, Power_Best, YX[:,1:], YX[:,0]
    
def TheilSenRegression(x,y, n_subsamples=None, random_state=42):
    """Calculate the Theil-Sen Linear Regression Model. This regression model is highly robust against
    outliers.
    
    Parameters
    ----------
    
    Returns
    -------
    line_x : ndarray
        Two values forming the (min(x), max(x))
    y_predict : ndarray
        Two values with the associated y values as predicted by the model
    coeff : float
        The coefficient of regression model
    intercept : float
        The intercept of the regression model
    r-value : float
        The r-value of the regression model
        
    Notes
    -----
    The p-value and standard error has not been implemented yet. Require bootstrapping, which is slow as
    the TheilSenRegression model is already slow to compute (~100 times slower then OLS).
    
    References
    ----------
    1) Theil-Sen Estimators in a Multiple Linear Regression Model, 2009 Xin Dang, Hanxiang Peng, 
    Xueqin Wang and Heping Zhang http://home.olemiss.edu/~xdang/papers/MTSE.pdf
    
    2) http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.TheilSenRegressor.html
    
    """
    
    #Initialise the Theil-Sen Linear Regression Model
    estimator = TheilSenRegressor(n_subsamples=n_subsamples, random_state=random_state)
    
    #Remove nan's in data
    x, y = antifinite((x,y), unpack=True)
    
    #Fit the regression model to data. N.B. add extra dimension [:,None] to x to calculate intercept.
    estimator.fit(x[:,None],y)
    
    #Predict the data from the model
    line_x = np.array([np.nanmin(x), np.nanmax(x)])
    y_predict = estimator.predict(line_x[:,None])
    
    #Calculate r-value
    r_value = power(R1to1(y, estimator.predict(x[:,None])), 0.5)
    
    return line_x, y_predict, estimator.coef_, estimator.intercept_, r_value

def HuberRegression(x,y, epsilon=1.36, package='statsmodel'):
    """
    Calculates the Huber linear regression model. Like the Theil-Sen 
    model, the Huber is resilient to outliers present in the data. 
    
    References
    ----------
    1) Peter J. Huber, Elvezio M. Ronchetti, Robust Statistics 
        Concomitant scale estimates, p.172
    2) http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.HuberRegressor.html
    """
    
    if Huber is False: 
        raise ImportError("[Warning] HuberRegression is not available in " \
                + "this version of python. :(")
    
    # Remove invalid data (nan's and inf's)
    x, y = antifinite((x,y), unpack=True)
        
    if package == 'sklearn':
        
        # Initialise the Theil-Sen Linear Regression Model
        estimator = HuberRegressor(epsilon=epsilon)
        
        #Fit the regression model to data. N.B. add extra dimension [:,None] to x to calculate intercept.
        estimator.fit(x[:,None],y)
        
        #Predict the data from the model
        line_x = np.array([np.nanmin(x), np.nanmax(x)])
        y_predict = estimator.predict(line_x[:,None])
        
        #Calculate r-value
        r_value = power(R1to1(y, estimator.predict(x[:,None])), 0.5)
        
        #return line_x, y_predict, estimator.coef_, estimator.intercept_, r_value
        return HuberRegressionResult(estimator.coef_[0], estimator.intercept_, r_value, np.nan, np.nan)
    
    elif package == 'statsmodel':
        
        #Initalise Huber model
        rlm_model = sm.RLM(y, sm.add_constant(x), M=sm.robust.norms.HuberT(epsilon))
        
        #fit model
        rlm_results = rlm_model.fit()
        
        #Calculate r-value
        r_value = power(R1to1(y, rlm_results.predict(sm.add_constant(x))), 0.5)
    
        return HuberRegressionResult(rlm_results.params[1], rlm_results.params[0], r_value, rlm_results.pvalues[1], rlm_results.bse[1])
    
def reg_m(y, x):
    """Multiple linear regression using statsmodel:
    
    y = dependent variable (1D array)
    x = independent variable(s) (2D array)
    
    https://stackoverflow.com/a/14971531"""

    ones = np.ones(len(x[0]))
    X = sm.add_constant(np.column_stack((x[0], ones)))
    for ele in x[1:]:
        X = sm.add_constant(np.column_stack((ele, X)))
    results = sm.OLS(y, X).fit()
    
    return results    

def rmse(predictions, targets):
    """Root Mean Square Error (RMSE) in easy form.
    
    Reference
    ---------
    https://stackoverflow.com/a/37861832/8765762
    """
    
    differences = predictions - targets                       #the DIFFERENCEs.

    differences_squared = differences ** 2                    #the SQUAREs of ^

    mean_of_differences_squared = differences_squared.mean()  #the MEAN of ^

    rmse_val = np.sqrt(mean_of_differences_squared)           #ROOT of ^

    return rmse_val                                           #get the ^

def R1to1(X, Y):
    """
    This function calculates the 1:1 R Squared value which can be 
    defined as the strength of the dependent variable to the 
    independent variable rather than the strenth of the linear 
    regression to to the dependent variable.
    
    Parameters
    ----------
    X : ndarray
        The independent variable such as the modelled data
    Y : ndarray
        The dependent variable such as the measured data
        
    NOTE: Requires numpy arrays to conduct element-wise operations!    
    """

    #Remove nan values from both model and measured datasets
    X,Y = antinan((X,Y), unpack=True)

    #Calculate the 1:1 R^2 from the remaining data
    SStot = np.sum((X - np.mean(X))**2)
    SSres = np.sum((X - Y)**2)
    Rsquared = 1 - SSres / SStot

    return Rsquared    

def RMSE(f, x):
    """
    Calculates the root mean square error of the data between x and y.
    
    Parameters
    ----------
    f, x : array_like
        Input data which must have the same dimensions. The index 
        values must also match up for the RMSE to be meaningful. i.e. 
        for all values of f there must be a corresponding x value. 
        
        N.B. f is defined as your model variable and x as your measured 
        variable.
    """

    f = np.asarray(f)
    x = np.asarray(x)

    f, x = antinan((f,x), unpack=True)

    # Check to see both inputs have equal size
    if f.size != x.size: 
        raise ValueError("[gu.RMSE] Inputs f and x must have equal " \
                         "dimensions. Got (%s, %s)" % (f.size, x.size))

    # Check neither inputs has a size 0
    if (f.size == 0) | (x.size == 0): 
        raise ValueError("[gu.RMSE] Inputs f and x must have data inside " \
                         "them. Go sizes of (%s, %s) for f and x " \
                         "respectively." % (f.size, x.size))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        return np.sqrt(np.sum((f - x)**2) / f.size)

def fstat(f, y, k=1, p_value=True):
    """Calculates the f-statistic from the fitted data, f and observed data, y.
    
    For a function, y = f(x,k) with n elements and k parameters,
    
    F = (SSM/DFM)/(SSE/DFE)
    
    SSM = Sum of squares for model
    SSE = Sum of squares for error
    DFM = Degrees of freedom for model
    DFE = Degrees of freedom for error
    
    DFM = k - 1
    DFE = n - k
    
    SSM = sum((y-mean(y))**2)
    SSE = sum((y-yf)**2)
    
    yf is the fitted value for y.
    
    Reference
    ---------
    http://facweb.cs.depaul.edu/sjost/csc423/documents/f-test-reg.htm
    https://stats.stackexchange.com/questions/254970/difference-between-ssm-and-ssr
    """
    
    #Calculate sum of square residuals
    SSE    = np.sum((y-f)**2)
    SSM = np.sum((y-np.nanmean(y))**2)
    
    DFE = y.size - k
    DFM = k - 1
    
    #Calculate F-statistic
    f_ratio = (SSM/DFM)/(SSE/DFE)
    
    #Calculate p-value
    p_value = 1 - sp.stats.f.cdf(f_ratio, DFM, DFE)
    
    return f_ratio, p_value
        
def running_mean(x, N):
    
    #Force N to be integer
    N = int(N)

    cumsum = np.cumsum(np.insert(x, 0, 0)) 
    return (cumsum[N:] - cumsum[:-N]) / N 
    
def moving_average(a, n=3):
    """Calculate the moving average of a 1D array with n being the number of terms to average over
    
    Reference
    ---------
    https://stackoverflow.com/a/14314054/8765762
    """
    
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

############################################################################
"""Kernal Density Functions

based on: https://jakevdp.github.io/blog/2013/12/01/kernel-density-estimation/
"""    

def kde_scipy(x, x_grid, bandwidth=0.2, **kwargs):
    """Kernel Density Estimation with Scipy"""
    
    #from kde import gaussian_kde
    from scipy.stats import gaussian_kde
    
    # Note that scipy weights its bandwidth by the covariance of the
    # input data.  To make the results comparable to the other methods,
    # we divide the bandwidth by the sample standard deviation here.
    kde = gaussian_kde(x, **kwargs)
    return kde.evaluate(x_grid)
    
def kde_statsmodels_u(x, x_grid, bandwidth=0.2, **kwargs):
    """Univariate Kernel Density Estimation with Statsmodels"""
    from statsmodels.nonparametric.kde import KDEUnivariate
    
    kde = KDEUnivariate(x)
    kde.fit(bw=bandwidth, **kwargs)
    return kde.evaluate(x_grid)
  
############################################################################
"""Descriptive Statistics"""
  
def Radar_Stats(Mag_Time, Mag_Height, Mag, Base_Time, Base, Base_Val=0):
    """Calculates the radar stats from the input product.
    
    You need to input the following datasets:
    
    Mag_Time
    Mag_Height
    Magnitude (e.g. dBZ)
    Base

    """
    
    if np.sum(np.isnan(Base)) != 1 or Base_Val == 0:
        cloud_R = np.where((Mag>-100)&(Mag_Time>=start)&(Mag_Time<=end)&(Mag_Height>=np.min(Base[np.where((Base_Time>=start)&(Base_Time<=end))])), Mag, np.nan); cloud_R = cloud_R[~np.isnan(cloud_R)]
        cloud_base = np.where((Base_Time>=start)&(Base_Time<=end), flatten(Base), np.nan); cloud_base = cloud_base[~np.isnan(cloud_base)]
    else:
        cloud_R = np.where((Mag>-100)&(Mag_Time>=start)&(Mag_Time<=end)&(Mag_Height>=Base_Val), Mag, np.nan); cloud_R = cloud_R[~np.isnan(cloud_R)]
        
        
    print("R Stats:", np.mean(cloud_R), np.median(cloud_R), np.std(cloud_R), np.max(cloud_R),np.min(cloud_R))
    print("dR Stats:", np.mean(cloud_dR), np.median(cloud_dR), np.std(cloud_dR), np.max(cloud_dR),np.min(cloud_dR))
    print("Base Stats:", np.mean(cloud_base), np.median(cloud_base), np.std(cloud_base), np.max(cloud_base),np.min(cloud_base))

def stats(array, extend=True, axis=None, output=False):
    """Outputs the standard statistics from the input array
    
    Parameters
    ----------
    array : ndarray
        The values you want to produce a statistic for
    extend : boolean, optional
        Specify outputting extra statistics. The extra ones are range and noise 
        (e.g. noise = median/std)
    axis : int, optional
        The axis you want to provide the stats. Default is None which averages as
        a single unit.
        
    Outputs
    -------
    stats : ndarray
        stats array containing the min, max, mean, median and std
    """
    
    #Error Control
    if array is None: 
        if extend is not False:
            return np.array([np.nan]*14)
        else:
            return np.array([np.nan]*5)
    
    # When array is not None, ensure it is numpy array
    array = np.atleast_1d(array)
    
    if extend:
        if array.size == 0:
            return np.array([np.nan]*14)
        else:
            if output is False:
                return DescriptiveStats_Extend(
                    np.size(array),
                    np.nanmin(array, axis=axis),
                    np.nanmax(array, axis=axis),
                    np.nanmean(array, axis=axis),
                    np.nanmedian(array, axis=axis),
                    np.nanstd(array, axis=axis),
                    np.nansum(array, axis=axis),
                    np.nanmax(array, axis=axis)-np.nanmin(array, axis=axis),
                    np.nanmedian(array, axis=axis)/np.nanstd(array, axis=axis),
                    np.nanpercentile(array, 5, axis=axis),
                    np.nanpercentile(array, 10, axis=axis),
                    np.nanpercentile(array, 25, axis=axis),
                    np.nanpercentile(array, 75, axis=axis),
                    np.nanpercentile(array, 90, axis=axis),
                    np.nanpercentile(array, 95, axis=axis))
            
            else:
                print("Descriptive Statistics")
                print("----------------------")
                print("Count = %s" % np.size(array))
                print("Minimum = %s" % np.nanmin(array, axis=axis))
                print("Maximum = %s" % np.nanmax(array, axis=axis))
                print("Mean = %s" % np.nanmean(array, axis=axis))
                print("Median = %s" % np.nanmedian(array, axis=axis))
                print("Standard Deviation = %s" % np.nanstd(array, axis=axis))
                print("Total = %s" % np.nansum(array, axis=axis))
                print("Range = %s" % (np.nanmax(array, axis=axis)-np.nanmin(array, axis=axis)))
                print("Noise = %s" % (np.nanmedian(array, axis=axis)/np.nanstd(array, axis=axis)))
                print("5th Percentile = %s" % np.nanpercentile(array, 5, axis=axis))
                print("10th Percentile = %s" % np.nanpercentile(array, 10, axis=axis))
                print("25th Percentile = %s" % np.nanpercentile(array, 25, axis=axis))
                print("75th Percentile = %s" % np.nanpercentile(array, 75, axis=axis))
                print("90th Percentile = %s" % np.nanpercentile(array, 90, axis=axis))
                print("95th Percentile = %s" % np.nanpercentile(array, 95, axis=axis))
                print("----------------------")
    else:
        if array.size == 0:
            np.array([np.nan]*5)
        else:
            if output is False:
                return DescriptiveStats(
                    np.size(array),
                    np.nanmin(array, axis=axis),
                    np.nanmax(array, axis=axis),
                    np.nanmean(array, axis=axis),
                    np.nanmedian(array, axis=axis),
                    np.nanstd(array, axis=axis))
            else:
                print("Descriptive Statistics")
                print("----------------------")
                print("Count = %s" % np.size(array))
                print("Minimum = %s" % np.nanmin(array, axis=axis))
                print("Maximum = %s" % np.nanmax(array, axis=axis))
                print("Mean = %s" % np.nanmean(array, axis=axis))
                print("Median = %s" % np.nanmedian(array, axis=axis))
                print("Standard Deviation = %s" % np.nanstd(array, axis=axis))

def iqr(array, percentile=75, axis=None):
    """Calculates the interquartile range of an array"""
            
    try:
        return np.subtract(*np.nanpercentile(array, [percentile,100-percentile], axis=axis))
    except TypeError:
        return np.nan

def lowess(x, y, f=2. / 3., iter=3):
    """lowess(x, y, f=2./3., iter=3) -> yest

    Lowess smoother: Robust locally weighted regression.
    The lowess function fits a non-parametric regression curve to a scatter plot.
    The arrays x and y contain an equal number of elements; each pair
    (x[i], y[i]) defines a data point in the scatter plot. The function returns
    the estimated (smooth) values of y.

    The smoothing span is given by f. A larger value for f will result in a
    smoother curve. The number of robustifying iterations is given by iter. The
    function will run faster with a smaller number of iterations.
    """
    
    n = len(x)
    r = int(np.ceil(f * n))
    h = [np.sort(np.abs(x - x[i]))[r] for i in range(n)]
    w = np.clip(np.abs((x[:, None] - x[None, :]) / h), 0.0, 1.0)
    w = (1 - w ** 3) ** 3
    yest = np.zeros(n)
    delta = np.ones(n)
    for iteration in range(iter):
        for i in range(n):
            weights = delta * w[:, i]
            b = np.array([np.sum(weights * y), np.sum(weights * y * x)])
            A = np.array([[np.sum(weights), np.sum(weights * x)],
                          [np.sum(weights * x), np.sum(weights * x * x)]])
            beta = sp.linalg.solve(A, b)
            yest[i] = beta[0] + beta[1] * x[i]

        residuals = y - yest
        s = np.median(np.abs(residuals))
        delta = np.clip(residuals / (6.0 * s), -1, 1)
        delta = (1 - delta ** 2) ** 2

    return yest

def SSR(x, y):
    """
    Calcualtes the sum of the square residuals between x and y
    """
    
    # Ensure data is an ndarray
    x = np.asarray(x)
    y = np.asarray(y)
    
    # Calculate sum of the squre residuals
    return np.nansum((x - y)**2)
    
def log_likelihood(model, endog):
    """
    Calcualtes the log-likelihood function from a one-dimensional set
    
    Parameters
    ----------
    model : 1D array_like
        A one-dimensional array_like of the model output. Size must
        match endog
    endog : 1D array_like
        A one-dimensional array_like of the data to create the model.
        (e.g. for a model y = m*x + c, endog is the y-data)
    """
    
    # Ensure data is an ndarray
    model = np.asarray(model)
    endog = np.asarray(endog)
    
    # Error checking
    if (model.ndim != 1) or (endog.ndim != 1):
        raise ValueError("log_likelihood requires the inputs to be 1 dimensional")
    
    if model.size != endog.size:
        raise ValueError("log_likelihood requires model and endog to have the same length")
    
    # Calcualte number of observations
    nobs = endog.size
    nobs2 = endog.size / 2.0

    # Calculate the mean and variance
    mean = np.nanmean(endog)
    variance = np.nanvar(endog)
    
    # Calculate sum of square residuals
    ssr = SSR(endog, model)
    
    # Calculate log-likelihood
    # return -nobs / 2 * (np.log(2 * np.pi) - np.log(variance)) - ssr / (2 * variance)        # Wikipedia method
    return -nobs2 * (np.log(2 * np.pi) + np.log(ssr / nobs) + 1)                        # Statsmodel method

def pearson_error(arr, n, alpha=0.05, output='lowhigh'):
    """
    Calculates the standard error of for the Pearson correlation 
    coefficient.
    
    References
    ----------
    https://en.wikipedia.org/wiki/Pearson_correlation_coefficient#Inference
    https://stats.stackexchange.com/a/73628
    """
    
    # Convert arr to an array
    arr = np.asarray(arr)
    
    # Convert n to an array if an array_like has been given
    if isarray(n):
        n = np.array(n, dtype=int)
    
    # Calculate z-score using alpha
    z_score = sp.stats.norm.ppf(1 - alpha / 2)

    # Calculate upper and lower confidence levels
    lower = np.tanh(np.arctanh(arr) - z_score / (n - 3)**0.5)
    upper = np.tanh(np.arctanh(arr) + z_score / (n - 3)**0.5)
    
    if output == 'lowhigh':
        return upper - lower
    elif output == 'errorbar':
        return np.array([arr - lower, upper - arr])
    elif output == 'confidence_intervals':
        return np.array([lower, upper])
    else:
        raise ValueError("Output option {0} is not supported. Choose either" \
                         "'lowhigh', 'errorbar' or 'confidence_intervals'".format(output))

def AIC(model, endog, k):
    """
    Calculate the Akaike Information Criterion
    
    Parameters
    ----------
    model : 1D array_like
        A one-dimensional array_like of the model output. Size must
        match endog
    endog : 1D array_like
        A one-dimensional array_like of the data to create the model.
        (e.g. for a model y = m*x + c, endog is the x-data)
    k : int
        The number of parameters used to define the model. This is the
        sum of the coefficients, intercept and variance. i.e. 
            
        k = c + i + v
    
    References
    ----------
    https://www.researchgate.net/post/What_is_the_AIC_formula
    https://stats.stackexchange.com/questions/87345/calculating-aic-by-hand-in-r
    https://www.statsmodels.org/dev/_modules/statsmodels/regression/linear_model.html#OLS.loglike
    https://en.wikipedia.org/wiki/Akaike_information_criterion#cite_note-19    
    """
    
    return 2 * k - 2 * log_likelihood(model, endog)

def AICc(model, endog, k):
    """ 
    Calcualtes the adjusted Akaike Information Criterion which is useful
    when the sample size is small [What defines small?]. This is used
    to avoid overfitting of the standard AIC diagnostic.
    
    Parameters
    ----------
    See AIC
    """
    
    return AIC(model, endog, k) + (2 * k**2 + 2 * k) / (np.size(endog) - k - 1)
        
def ensemble(xdata, ydata, bins, average=('mean', 'median'), method='ma', 
        undersample=False, slim=False, usestride=False, 
        confidence_intervals='standard', x_err=False, unpack=False):
    """Averages the ydata set by binning the xdata set into bins of equal elements.
    
    Parameters
    ----------
    xdata : 1D array
        The xdata set to represent the midpoints of each bin
    ydata : 1D array
        The ydata set that is used for averaging
    bins : int or list or ndarray
        The number of bins you want to average the dataset over. Supply bin with a list
        or ndarray when you want to manual choose the bin locations.
    average : str or array_like of string, optional
        Specify if you want to average the data placed in each bin using the median or
        mean. Option to specify a 2 element array to specifiy the average type for the
        xdata and ydata;
        
                        average = (xdata_type, ydata_type)
                                = ('mean', 'median')
    method : str or int, optional
        Specify if you should average over unique elements only (default = 'unique') or
        whether you should roll the bins over all elements ('ma'). If an integer is 
        specified the number will be used to select the number of elements to roll over.
        Final method is 'bootstrap' which slowly increases the number of data elements
        in each bin until the entire population is in one bin. Then the reverse occurs
        when the first elements are removed one by one until only one element remains.
    undersample : boolean, optional
        Specify if you should average the areas of data that are under sampled. Default
        is false and therefore does not average the under sample data. When set to True
        the number of elements are in the bins near the boundary are subsequently 
        reduced until only a 1 element bin exists.
    confidence_intervals : str, optional
        Specify how you want to define the errors of the averaged data. All confidence
        intervals are given as 95th percentiles. Options are,
        
        'standard' : use the standard error formula. If median averaging is used the
                     approximation is used to correct for the sample. SE = 1.253*SE
        'bootstrap' : use a bootstrapping algorithm to estimate the confidence
                      intervals. N.B. this method is very slow.
        
    Examples
    --------
    If we define an example dataset and the number of bins we want to place our data
    in:
    
    a = np.arange(100)
    bins = 10
    
    We can set the mode to either true or false depending on if we want each element
    of our dataset, a, to be place in one bin or several.
    
    If mode is False, then the bins will have the structure as follows,
        
                            0-9, 10-19, 20-29 ... etc. 
                            
    and therefore each element can only exist in one unique bin.
    
    If mode is True, then the bins will have the structure as follows,
    
                            0-9, 1-10, 2-11, 3-12 ... etc.
                            
    and therefore each element can be in multiple bins. This method is used to reduce the
    discontinuities in the data while providing a greater degree of resolution in the
    averaging.
    
    Issues
    ------
    1)    The broadcasting of the xdata is done separately to the broadcasting of the ydata
        to reduce the chance for MemoryError when the size of the xdata and ydata are extremely
        large. Depending on the bin number and especially when mode is set to True, the size of  
        the broadcast array can be huge. This is the only disadvantage of performing the 
        averaging using the ensemble function. Otherwise, this is by far the superior method in 
        terms of processing time and simplicity.
    
    Solved Issues
    -------------
    2)    When using a small bin number the data will fail to represent the under sampled regions
        very well. This is shown when mode is True (i.e. we use a rolling window with a 
        predetermined number of elements in each bin) where a lower bin number has more elements
        available for averaging increasing the representation. This only provides a good 
        representation for the highly sampled areas of data. A potential solution for this issue
        is to downscale the number of elements selected near the boundary thus making it possible
        to represent these areas of data. It must be stressed in either report writing or to the
        user that these downscaled areas must be taken with caution as they have not been averaged
        in the same manner as the well sample areas of the dataset.
        
    3) Now supports correct assessment of confidence intervals when the median statistic is used.
    
    """

    # Sort data by ascending xdata
    mask = np.argsort(xdata, kind='mergesort')
    ydata = ydata[mask].astype(float)
    xdata = xdata[mask].astype(float)
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        
        if method !=  'bootstrap':
            if isinstance(bins, int):
            
                #Constants
                elems = int(np.floor(ydata.size/bins))

                if undersample is False:
                    """data_bin_x and data_bin_y are FLOAT ndarrays so element-wise operations
                    CAN be performed succinctly (e.g. np.nanmean, np.nanmedian and np.nanstd)
                    (i.e. THIS METHOD IS FAST)
                    """
                    if usestride is False:
                        xdata_average = average if isinstance(average, str) else average[0]
                        
                        #Calculate MidPoint
                        data_bin_x = broadcast(xdata, elems, elems) if method == 'unique' else broadcast(xdata, elems, method)
                        MidPoints = np.nanmedian(data_bin_x, axis=1) if xdata_average == 'median' else np.nanmean(data_bin_x, axis=1)
                        
                        #Calculate Boundaries
                        Bin_Size = np.tile(data_bin_x.shape[1], data_bin_x.shape[0]) if slim is False else np.zeros(MidPoints.size)
                        #Boundaries = [MidPoints[0], MidPoints[-1]]
                        
                        # Calculate confidence intervals for xdata
                        if x_err:
                            if confidence_intervals == 'standard':
                                xdata_se = 1.96*np.nanstd(data_bin_x, axis=1)/np.sqrt(elems) if xdata_average == 'mean' else 1.96*(1.253*np.nanstd(data_bin_x, axis=1)/np.sqrt(elems))
                            elif confidence_intervals == 'bootstrap':
                                xdata_se = boot.ci(data_bin_x.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if xdata_average == 'mean' else boot.ci(data_bin_x.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                            else:
                                raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                    
                        del data_bin_x
                        
                        ydata_average = average if isinstance(average, str) else average[1]
                        
                        #Calculate Average
                        data_bin_y = broadcast(ydata, elems, elems) if method == 'unique' else broadcast(ydata, elems, method)
                        Average = np.nanmedian(data_bin_y, axis=1) if ydata_average == 'median' else np.nanmean(data_bin_y, axis=1)
                        
                        #Calculate Confidence Limit
                        if confidence_intervals == 'standard':
                            SE = 1.96*np.nanstd(data_bin_y, axis=1)/np.sqrt(elems) if ydata_average == 'mean' else 1.96*(1.253*np.nanstd(data_bin_y, axis=1)/np.sqrt(elems))
                        elif confidence_intervals == 'bootstrap':
                            SE = boot.ci(data_bin_y.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if ydata_average == 'mean' else boot.ci(data_bin_y.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                        else:
                            raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                    
                    else:
                        xdata_average = average if isinstance(average, str) else average[0]
                        
                        #Calculate MidPoint
                        data_bin_x = stride(xdata, elems, elems) if method == 'unique' else stride(xdata, elems, method)
                        MidPoints = np.nanmedian(data_bin_x, axis=1) if xdata_average == 'median' else np.nanmean(data_bin_x, axis=1)
                        
                        #Calculate Boundaries
                        Bin_Size = np.tile(data_bin_x.shape[1], data_bin_x.shape[0]) if slim is False else np.zeros(MidPoints.size)

                        # Calculate confidence intervals for xdata
                        if x_err:
                            if confidence_intervals == 'standard':
                                xdata_se = 1.96*np.nanstd(data_bin_x, axis=1)/np.sqrt(elems) if xdata_average == 'mean' else 1.96*(1.253*np.nanstd(data_bin_x, axis=1)/np.sqrt(elems))
                            elif confidence_intervals == 'bootstrap':
                                xdata_se = boot.ci(data_bin_x.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if xdata_average == 'mean' else boot.ci(data_bin_x.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                            else:
                                raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                    
                        del data_bin_x
                        
                        ydata_average = average if isinstance(average, str) else average[1]
                        
                        #Calculate Average
                        data_bin_y = stride(ydata, elems, elems) if method == 'unique' else stride(ydata, elems, method)
                        Average = np.nanmedian(data_bin_y, axis=1) if ydata_average == 'median' else np.nanmean(data_bin_y, axis=1)
                        
                        #Calculate Confidence Limit
                        if confidence_intervals == 'standard':
                            SE = 1.96*np.nanstd(data_bin_y, axis=1)/np.sqrt(elems) if ydata_average == 'mean' else 1.96*(1.253*np.nanstd(data_bin_y, axis=1)/np.sqrt(elems))
                        elif confidence_intervals == 'bootstrap':
                            SE = boot.ci(data_bin_y.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if ydata_average == 'mean' else boot.ci(data_bin_y.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                        else:
                            raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                
                elif undersample is True:
                    """data_bin_x and data_bin_y are OBJECT ndarrays so element-wise operations
                    CANNOT be performed succinctly. Therefore, we have to calculate the mean, 
                    median and std for each bin in tern (i.e. THIS METHOD IS SLOW).
                    
                    Boundaries are given for when the data becomes under-sampled and an equal 
                    number of elements per bin does not hold true.
                    """
                    
                    if usestride is False:
                        xdata_average = average if isinstance(average, str) else average[0]
                        
                        #Calculate MidPoint
                        data_bin_x = broadcast(xdata, elems, elems, undersample=True) if method == 'unique' else broadcast(xdata, elems, method, undersample=True)
                        MidPoints = np.nanmedian(data_bin_x, axis=1) if xdata_average == 'median' else np.nanmean(data_bin_x, axis=1)

                        #Calculate Under-sampled Boundaries
                        Bin_Size = np.array([len(antinan(data)) for data in data_bin_x], dtype=int) if slim is False else np.zeros(MidPoints.size)
                        #MidPoints_Full = MidPoints[Bin_Size == np.max(Bin_Size)]
                        #Boundaries = np.array([MidPoints_Full[0], MidPoints_Full[-1]], dtype=float)
                        
                        # Calculate confidence intervals for xdata
                        if x_err:
                            if confidence_intervals == 'standard':
                                xdata_se = 1.96*np.nanstd(data_bin_x, axis=1)/np.sqrt(elems) if xdata_average == 'mean' else 1.96*(1.253*np.nanstd(data_bin_x, axis=1)/np.sqrt(elems))
                            elif confidence_intervals == 'bootstrap':
                                xdata_se = boot.ci(data_bin_x.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if xdata_average == 'mean' else boot.ci(data_bin_x.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                            else:
                                raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                    
                        del data_bin_x
                        
                        ydata_average = average if isinstance(average, str) else average[1]
                        
                        #Calculate Average
                        data_bin_y = broadcast(ydata, elems, elems, undersample=True) if method == 'unique' else broadcast(ydata, elems, method, undersample=True)
                        Average = np.nanmedian(data_bin_y, axis=1) if ydata_average == 'median' else np.nanmean(data_bin_y, axis=1)
                        
                        #Calculate Confidence Limit
                        if confidence_intervals == 'standard':
                            SE = 1.96*np.nanstd(data_bin_y, axis=1)/np.sqrt(elems) if ydata_average == 'mean' else 1.96*(1.253*np.nanstd(data_bin_y, axis=1)/np.sqrt(elems))
                        elif confidence_intervals == 'bootstrap':
                            SE = boot.ci(data_bin_y.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if ydata_average == 'mean' else boot.ci(data_bin_y.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                        else:
                            raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                    
                    else:
                        xdata_average = average if isinstance(average, str) else average[0]
                        
                        #Calculate MidPoint
                        data_bin_x = stride(xdata, elems, elems, undersample=True) if method == 'unique' else stride(xdata, elems, method, undersample=True)
                        MidPoints = np.nanmedian(data_bin_x, axis=1) if xdata_average == 'median' else np.nanmean(data_bin_x, axis=1)
                        #print("MIDPOINTS DONE")
                        #Calculate Under-sampled Boundaries
                        Bin_Size = np.array([len(antinan(data)) for data in data_bin_x], dtype=int) if slim is False else np.zeros(MidPoints.size)
                        #MidPoints_Full = MidPoints[Bin_Size == np.max(Bin_Size)]
                        #Boundaries = np.array([MidPoints_Full[0], MidPoints_Full[-1]], dtype=float)
                        
                        # Calculate confidence intervals for xdata
                        if x_err:
                            if confidence_intervals == 'standard':
                                xdata_se = 1.96*np.nanstd(data_bin_x, axis=1)/np.sqrt(elems) if xdata_average == 'mean' else 1.96*(1.253*np.nanstd(data_bin_x, axis=1)/np.sqrt(elems))
                            elif confidence_intervals == 'bootstrap':
                                xdata_se = boot.ci(data_bin_x.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if xdata_average == 'mean' else boot.ci(data_bin_x.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                            else:
                                raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                    
                        del data_bin_x
                        
                        ydata_average = average if isinstance(average, str) else average[1]
                        
                        #Calculate Average
                        data_bin_y = stride(ydata, elems, elems, undersample=True) if method == 'unique' else stride(ydata, elems, method, undersample=True)
                        Average = np.nanmedian(data_bin_y, axis=1) if ydata_average == 'median' else np.nanmean(data_bin_y, axis=1)
                        
                        #Calculate Confidence Limit
                        if confidence_intervals == 'standard':
                            SE = 1.96*np.nanstd(data_bin_y, axis=1)/np.sqrt(elems) if ydata_average == 'mean' else 1.96*(1.253*np.nanstd(data_bin_y, axis=1)/np.sqrt(elems))
                        elif confidence_intervals == 'bootstrap':
                            SE = boot.ci(data_bin_y.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if ydata_average == 'mean' else boot.ci(data_bin_y.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                        else:
                            raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                    
            else:
                #Prerequisites
                bins = np.array(bins)
                MidPoints = bins
                ydata_average = average if isinstance(average, str) else average[1]
                
                #Digitise data
                data_mask = digitize(xdata, bins)
                #Bin_Size = np.bincount(data_mask)
                max_size = np.max(np.bincount(data_mask))
                
                #Group data into rows and pad to make rows have equal length
                data_bin_y = np.array([np.pad(ydata[data_mask == i], (0, max_size - ydata[data_mask == i].size), 'constant', constant_values=np.nan) for i, bin in enumerate(bins)], dtype=np.float64)
                Bin_Size = np.array([np.sum(data_mask == i) for i, bin in enumerate(bins)], dtype=np.int32)
                
                #Average data
                if ydata_average == 'median':
                    Average = np.nanmedian(data_bin_y, axis=1) 
                elif ydata_average == 'mean':
                    Average = np.nanmean(data_bin_y, axis=1)
                elif ydata_average == 'sum':
                    Average = np.nansum(data_bin_y, axis=1)

                # Calculate Confidence Limit
                if confidence_intervals == 'standard':
                    SE = 1.96 * np.nanstd(data_bin_y, axis=1) / np.sqrt(np.size(data_bin_y, axis=1)) if ydata_average == 'mean' else 1.96 * (1.253 * np.nanstd(data_bin_y, axis=1) / np.sqrt(np.size(data_bin_y, axis=1)))
                elif confidence_intervals == 'bootstrap':
                    SE = boot.ci(data_bin_y.T, lambda x: np.mean(x, axis=0)).ptp(axis=0) if ydata_average == 'mean' else boot.ci(data_bin_y.T, lambda x: np.median(x, axis=0)).ptp(axis=0)
                else:
                    raise ValueError("[gu.ensemble] confidence_intervals was incorrectly specified. We got %s. Available options are: 'standard', 'bootstrap'" % confidence_intervals)
                    
        else:
            """Groups the data using bootstrapping method"""
            
            #Constants
            elems = ydata.size/bins
            xdata_average = average if isinstance(average, str) else average[0]
            ydata_average = average if isinstance(average, str) else average[1]
            
            #Format x-data
            data_bin_x = stride(xdata, xdata.size, 1, undersample=True)[elems-1::elems]
            MidPoints = np.nanmedian(data_bin_x, axis=1) if xdata_average == 'median' else np.nanmean(data_bin_x, axis=1)
            
            #Format y-data
            data_bin_y = stride(ydata, ydata.size, 1, undersample=True)[elems-1::elems]
        
            Average = np.nanmedian(data_bin_y, axis=1) if ydata_average == 'median' else np.nanmean(data_bin_y, axis=1)
            SE = np.nanstd(data_bin_y, axis=1)/np.sqrt(ydata.size) if ydata_average == 'mean' else boot.ci(data_bin_y, lambda x: np.median(x, axis=1)).ptp(axis=1)
        
            #Force slim to be True
            slim = True
    
    if unpack is False:
        if slim is False:
            return np.vstack((MidPoints, Average, xdata_se, SE, Bin_Size, data_bin_y.T)).T if x_err else np.vstack((MidPoints, Average, SE, Bin_Size, data_bin_y.T)).T
        else:
            return np.vstack((MidPoints, Average, xdata_se, SE)).T if x_err else np.vstack((MidPoints, Average, SE)).T
    else:
        if slim is False:
            return MidPoints, Average, xdata_se, SE, Bin_Size, data_bin_y.T if x_err else MidPoints, Average, SE, Bin_Size, data_bin_y.T
        else:
            if x_err:
                return MidPoints, Average, xdata_se, SE 
            else:
                return MidPoints, Average, SE
    


import sys
import numpy as np
import pandas as pd
import warnings, sys
from datetime import date, datetime, timedelta
from datetime import time as dttime

from .system import isarray

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
	xrange = range

def time32(when='now', dtype='str'):
    """Output the time in different formats
    
    Parameters
    ----------
    when : str, optional
        Specify the time you want to configure. Currently only 'now' is an
        available option.
    dtype : str, optional
        Specify the data type you want the time to be given in. The current
        options are 'str', 'datetime' and 'datetime64' which corresponds to
        a string, python datetime and numpy datetime object output.
        
    Output
    ------
    time : dtype
        Returns the time in various types depending on the input data type.
        
        str : Return hh:mm:ss string representation of the current time
        
        datetime : Return the python datetime object
        
        datetime64 : Return the numpy datetime object
        
    """
    
    if when == 'now':
        t = datetime.now()
    
    if dtype == 'str':
        return t.time().strftime("%H:%M:%S")
    elif dtype == 'datetime':
        return t.time()
    elif dtype == 'datetime64':
        return np.datetime64(t).astype('datetime64[s]')
    else:
        raise warnings.warn('[gu.time]: The specified dtype must be either "str", "datetime" or "datetime64". We got %s' % (dtype), UserWarning, stacklevel=2)

def datetime32(when='now', dtype='str'):
    """Output the date and time in different formats
    
    Parameters
    ----------
    when : str, optional
        Specify the time you want to configure. Currently only 'now' is an
        available option.
    dtype : str, optional
        Specify the data type you want the time to be given in. The current
        options are 'str', 'datetime' and 'datetime64' which corresponds to
        a string, python datetime and numpy datetime object output.
        
    Output
    ------
    time : dtype
        Returns the time in various types depending on the input data type.
        
        str : Return hh:mm:ss string representation of the current time
        
        datetime : Return the python datetime object
        
        datetime64 : Return the numpy datetime object
        
    """
    
    if when == 'now':
        t = datetime.now()
    
    if dtype == 'str':
        return t.strftime("%Y/%m/%d %H:%M:%S")
    elif dtype == 'datetime':
        return t
    elif dtype == 'datetime64':
        return np.datetime64(t).astype('datetime64[s]')
    else:
        raise warnings.warn('[gu.time]: The specified dtype must be either "str", "datetime" or "datetime64". We got %s' % (dtype), UserWarning, stacklevel=2)
            
def Excel_to_Python_Date(excel_time, expand=None, format=None, strip=True):
    """Converts Excel formatted time into a python format
    
    Parameters
    ----------
    time : numpy array
        1 dimensional time series array attached to data stream to determine the timing a step
        was detected
    expand : boolean, optional
        Outputs an expanded array of time variants over the broad datetime format. This is
        useful when requiring year day or fraction hours of the day
    format : str, optional
        if the excel string is in a different format then you can specify this here
    strip : boolean, optional
        strips the time component from datetime objects while leaving them intact
    
    Returns
    -------
    Python_Datetime : numpy array, float
        The date and time of each timestamp in datetime python format
    Python_Date : numpy array, float
        The date of each timestamp in datetime python format
    Python_Year : numpy array, int, optional
        The year at each time step
    Python_YD : numpy array, int, optional
        The day of the year at each timestep
    Python_Hour : numpy array, float, optional
        The fraction hour of each timestep assuming time has been given.
    """

    Python_Datetime = np.zeros(len(excel_time), dtype=object)
    if format is not None: #Bespoke method
        for i in xrange(len(excel_time)):
            try:
                Python_Datetime[i] = datetime.strptime(excel_time[i], format)
            except TypeError:
                continue
    else: #Standard Method
        Python_Datetime = np.array([datetime.strptime(date, '%Y-%m-%d %H:%M:%S') for date in excel_time], dtype=object)

    Python_Datetime    = Python_Datetime[Python_Datetime != 0]
    
    if expand is not None:
        #Convert Python_Datetime to decimal hour format
        Python_Date = np.zeros_like(Python_Datetime)
        Python_Year = np.zeros_like(Python_Datetime)
        Python_YD = np.zeros_like(Python_Datetime)
        Python_H = np.zeros_like(Python_Datetime)
        for i in xrange(len(Python_Datetime)):
            Python_Date[i] = Python_Datetime[i].date()
            Python_Year[i] = Python_Datetime[i].year
            Python_YD[i]     = Python_Datetime[i].timetuple().tm_yday
            Python_H[i] = toHourFraction(Python_Datetime[i])
    
    if strip == True:
        Python_Datetime = toDateOnly(Python_Datetime)
    
    return Python_Datetime if expand is None else (Python_Datetime, Python_Date, Python_Year, Python_YD, Python_H)
        
def toHourFraction(date):
    """Converts time into fractional hours between 0-24
    
    Parameters
    ----------
    date : numpy array
        1 dimensional time series array to convert

    Returns
    -------
    hour_frac : numpy array, float
        The hour fraction for each element of array 'date'
    """
    
    sec = (date-datetime.combine(date.date(), dttime(0,0,0))).total_seconds()
    hour_frac = sec/3600
    return hour_frac

def toDateOnly(dates, dtype='datetime'):
    """Converts an array of datetime objects and removes the time
    component keeping the datetime object intact. Similar to doing
    just <datetime_object>.date() but makes it comparable to other
    dates"""
    
    if dtype == 'datetime':
        if isinstance(dates, datetime): #Not an array of dates
            return datetime.combine(dates, dttime(0,0,0))
        else:
            return np.array([datetime.combine(dates[i], dttime(0,0,0)) for i in xrange(len(dates))], dtype=object)
    elif dtype == 'date':
        if isinstance(dates, datetime): #Not an array of dates
            return datetime.date()
        else:
            return np.array([DATETIME.date() for DATETIME in dates], dtype=object)
    else:
        raise ValueError("[toDateOnly] dtype input needs to be either 'datetime' or 'date'. We got %s" % dtype)
        
def toDatetime(date_val):
    """converts datetime.date() object into a datetime() object"""
    
    if isinstance(date_val, date):
        return datetime.fromordinal(date_val.toordinal())
    elif isinstance(date_val, np.ndarray):
        if date_val.dtype == 'O':
            return np.array([datetime.fromordinal(date2change.toordinal()) for date2change in date_val], dtype=object)
        else:
            sys.exit("[Error: toDatetime] Incorrect dtype")
    else:
        sys.exit("[Error: toDatetime] Incorrect data type")
        
def toDatetime64(arg, format=None, dtype='datetime64[us]'):
    """Converts datetime objects to numpy datetime64. Useful for converting entire array to a different datetime
    dtype.
    
    Parameters
    ----------
    arg : numpy.datetime64, datetime.datetime, or array_like
        
    dtype : str
        The numpy datetime64 dtype to convert the inputs to. Available options are dtype='datetime[?]' 
        where ? can include ['as', 'fs', 'ps', 'ns', us', 'ms', 's', 'm', 'h', 'D', 'W', 'M', 'Y'].
        
    Output
    ------
    array : numpy.datetime64 or np.ndarray
        Outputs all elements in arg either as a numpy.datetime64 or np.ndarray depending on whether the arg was
        array_like or not. The output will have the dtype as specified using the dtype parameter."""
    
    #Error checking
    if not isinstance(dtype, str): raise ValueError("[gu.toDatetime64] 'dtype' parameter was not specified correctly. \
        We got %s. Available options are dtype='datetime[?]' where ? can include ['as', 'fs', 'ps', 'ns',\
        us', 'ms', 's', 'm', 'h', 'D', 'W', 'M', 'Y']." % dtype)
    if not 'datetime' in dtype: raise ValueError("[gu.toDatetime64] 'dtype' parameter was not specified correctly. \
        We got %s. Available options are dtype='datetime[?]' where ? can include ['as', 'fs', 'ps', 'ns',\
        us', 'ms', 's', 'm', 'h', 'D', 'W', 'M', 'Y']." % dtype)
    
    #if arg is array_like
    if isarray(arg):
        #if arg is an array, then convert to numpy
        arg = np.asarray(arg)
        
        #Convert to numpy datetime64
        if isinstance(arg[0], date):
            return arg.astype('datetime64[D]').astype(dtype)
        elif isinstance(arg[0], str):
            if format is None:
                return np.array(pd.to_datetime(arg), dtype=np.datetime64)
            else:
                return np.array(pd.to_datetime(arg, format=format), dtype=np.datetime64)
        else:
            return arg.astype(dtype)
    
    #if arg is a numpy datetime64 object
    elif np.issubdtype(arg, np.datetime64):
        return arg.astype(dtype)
    
    #if arg is a datetime.datetime object
    elif isinstance(arg, datetime):
        return np.datetime64(arg).astype('datetime64[D]').astype(dtype)
    
    #if arg is a datetime.date object
    elif isinstance(arg, date):
        return np.datetime64(arg).astype(dtype)
    
    #Anything else
    else:
        #return None
        raise ValueError("[gu.toDatetime64] 'arg' parameter was not specified correctly. We got %s. arg must be \
            array_like or either datetime.datetime, datetime.date or np.datetime64")
    
def roundTime(dt=None, roundTo=60):
   """Round a datetime object to any time laps in seconds
   
   Parameters
   ----------
   dt : datetime object, optional
        The datetime you want to round. Default uses current time.
   roundTo : int, optional
        Closest number of seconds to round to, default 1 minute.
   
   Author and References
   ---------------------
   Thierry Husson 2012 - "Use it as you want but don't blame me."
   https://stackoverflow.com/a/10854034/8765762
   
   """
   
   if dt is None: dt = datetime.now()
   seconds = (dt.replace(tzinfo=None) - dt.min).seconds
   rounding = (seconds+roundTo/2) // roundTo * roundTo
   
   return dt + timedelta(0,rounding-seconds,-dt.microsecond)

def hours2dt(hours_since_epoch, epoch, dtype='datetime64'):
    """Converts an array of values representing the number of hours since a common point
    known as an epoch.
    
    Parameters
    ----------
    hours_since_epoch : ndarray
        An array of the hours since the epoch date
    epoch : str, datetime, datetime64
        The specific datetime you want to start from. The format for epoch needs to
        work with numpy datetime64.
    dtype : str, optional
        Specify if you want the output to be in numpy datetime or native "python"
        datetime format.
        
    Example
    -------
    >>> hours_since_epoch = np.linspace(0,24,86400)
    >>> epoch = datetime(1992,8,17)
    
    >>> hours2dt(hours_since_epoch, epoch)
    array([datetime(1992,8,17,0,0,0), datetime(1992,8,17,0,0,1) ...])
    
    Notes
    -----
    Function is fast if you specify dtype='datetime64' and leave it in numpy format.
        
    Reference
    ---------
    https://codereview.stackexchange.com/a/77662/151534
    https://stackoverflow.com/a/13704307/8765762
    """
    
    seconds = np.around(hours_since_epoch * (60*60))
    if dtype == 'datetime64':
        return np.datetime64(epoch) + seconds.astype('timedelta64[s]')
    elif dtype == 'datetime':
        return np.array((np.datetime64(epoch) + seconds.astype('timedelta64[s]')).astype(datetime), dtype=object)

def dt2hours(array_of_datetimes, epoch, checkback=False):
    """Converts an array of datetime objects to hours since epoch. Reverse of hours2dt.
    
    epoch can be either a single datetime or an ndarray of datetime matching the dimensions
    of array_of_datetimes exactly.
    
    np.datetime64('NaT') are converted to np.nan on output.
    
    checkback parameter is useful when blindly converting an array without knowing if the 
    input array contains any datetimes. outputs True if datetimes were found, else False"""
    
    if isinstance(array_of_datetimes, np.ndarray):

        #check for float and int dtypes
        dtypes = ['float', 'float32', 'float128', 'int']
        if np.any(np.in1d(dtypes, array_of_datetimes.dtype)):
            if checkback is False:
                return array_of_datetimes  
            else:
                return array_of_datetimes, False, array_of_datetimes.dtype
        else:
            if not isinstance(epoch, np.ndarray): epoch = np.datetime64(epoch) 
            if array_of_datetimes.dtype == 'O': #Object array
                try:
                    array_of_datetimes = array_of_datetimes.astype('datetime64[s]')
                    
                    #Find all np.nat values in datetime
                    mask = isnat(array_of_datetimes)
                    
                    #Convert to hours
                    hours = (array_of_datetimes - epoch.astype('datetime64[s]')).astype('timedelta64[s]').astype(float)/3600
                    
                    #Convert original np.nat values to np.nan
                    hours[mask] = np.nan
                
                except ValueError: #Issue can arise if input array is object but does not contain datetimes.
                    if checkback is False:
                        return np.array(array_of_datetimes, dtype=type(array_of_datetimes[0]))  
                    else:
                        return np.array(array_of_datetimes, dtype=type(array_of_datetimes[0])), False, np.array(array_of_datetimes, dtype=type(array_of_datetimes[0])).dtype
                
                if checkback is False:
                    return hours  
                else:
                    return hours, True, 'datetime'
            else:
                #Find all np.nat values in datetime
                mask = isnat(array_of_datetimes)
                
                #Convert to hours
                hours = (array_of_datetimes.astype('datetime64[s]') \
                    - epoch.astype('datetime64[s]')).astype('timedelta64[s]').astype(np.float64) \
                    / 3600
                
                #Convert original np.nat values to np.nan
                hours[mask] = np.nan
                
                if checkback is False:
                    return hours  
                else:
                    return hours, True, 'datetime64'
    
    elif isinstance(array_of_datetimes, np.datetime64):
        
        array_of_datetimes = np.array(array_of_datetimes)
        
        #Convert to hours
        hours = (array_of_datetimes.astype('datetime64[s]') \
            - epoch.astype('datetime64[s]')).astype('timedelta64[s]').astype(np.float64) \
            / 3600
        
        if checkback is False:
            return hours 
        else:
            return hours, True, 'datetime64'
    
    else:
        raise ValueError("{Error} dt2hours requires an ndarray as input not a single value")

def isnat(your_datetime):
    """Determines if a numpy datetime64 array has any "not a time" values inside
    
    Reference
    ---------
    https://stackoverflow.com/a/42103441/8765762
    
    """
    
    nat_as_integer = np.datetime64('NAT').view('i8')
    
    dtype_string = str(your_datetime.dtype)
    if 'datetime64' in dtype_string or 'timedelta64' in dtype_string:
        return your_datetime.view('i8') == nat_as_integer
    
    return np.array([], dtype=bool)  # it can't be a NaT if it's not a dateime

def isdatetime(arr):
    """
    Determines whether array contains datetime elements.
    """

    arr = np.atleast_1d(arr)

    # First, test whether all values are datetime compatible (Quick-Easy method)
    try:
        arr.view('datetime64[s]')
        return np.full(arr.size, True, dtype=bool)
    except TypeError:
        # Second, test each element in turn

        arr_bool = np.full(arr.size, False, dtype=bool)
        for i, val in enumerate(arr):
            if isinstance(val, datetime) \
                or isinstance(val, date) \
                or isinstance(val, dttime) \
                or isinstance(val, np.datetime64) \
                or isinstance(val, pd.Timestamp):

                    arr_bool[i] = True

        return arr_bool

def datetime64range(start, end, units, interval=1):#, elements=None):
    """Creates a numpy datetime64 array between start and end with the option to 
    specifiy the time units and interval
    
    Parameters
    ----------
    start : datetime or datetime64
        The datetime you want to start from.
    end : datetime or datetime64
        same as above but for the end time
    units : str, optional
        This specifies the type of time or date units you want. 
        Go to https://docs.scipy.org/doc/numpy-1.13.0/reference/arrays.datetime.html#datetime-units
        for specifications for units.
    interval : float or int, optional
        How many intervals do you want.
        
    """
    
    np.arange(start, end, timedelta(days=1)).astype(datetime)
    
    #Ensure datetime64 compliance
    start = np.datetime64(start)
    end = np.datetime64(end)
    
    #If user has not specified interval nor elements we select interval as 1
    #if interval is None and elements is None: interval = 1
    
    #if interval is not None:
    return np.arange(start.astype('datetime64[' + units + ']'), end.astype('datetime64[' + units + ']'), np.timedelta64(interval, units), dtype='datetime64[' + units + ']').astype(start.dtype)
    #elif elements is not None:
    #    return np.linspace(start.astype('datetime64[' + units + ']'), end.astype('datetime64[' + units + ']'), np.timedelta64(elements, units), dtype='datetime64[' + units + ']').astype(start.dtype)

def DatetimeFormat(date, format, check=False):
    """Formats and checks the validity of a datetime
    format as specified by the user.
    
    Parameters
    ----------
    date : str or list
        the datetime string you want to check for being 
        formatted in the correct format.
    format : str
        the format for the datetime to be formatted in.
        E.g. %d/%m/%Y_%H:%M:%S
        Check http://strftime.org/ if you're unsure.
    check : bool, optional
        whether to only check the format and return
        boolean indicators."""
    
    if date is None: 
        warnings.warn("[gu.DatetimeFormat] date input is None")
        return None
    
    if isinstance(date, str):
        if check is True:
            try:
                datetime.strptime(date, format)
                return True
            except ValueError:    
                return False
        else:    
            try:
                return datetime.strptime(date, format)
            except ValueError:    
                return np.nan
    else:
        Python_Datetime = np.zeros(len(date), dtype=object)
        for i in xrange(len(date)):
            try:
                Python_Datetime[i] = datetime.strptime(date[i], format)
            except TypeError:
                continue
                
        return Python_Datetime    

def ensure_strftime(input_date, format, yearfirst=True):
    """Ensures that the date input is in a string format which also matches the
    specified format"""
    
    from dateutil.parser import parse
    
    #Make sure format is in string format
    if not isinstance(format, str):
        raise ValueError("[gu.ensure_strftime]: Format dtype must be string. We got %s" % (type(format)))
    
    if isinstance(input_date, datetime) or isinstance(input_date, date):
        return input_date.strftime(format)
    elif isinstance(input_date, np.datetime64):
        return input_date.astype('datetime64[s]').astype(datetime).strftime(format)
    elif isinstance(input_date, str):
        if format.rfind("Y") + format.rfind("d") >= 0:
            if format.rfind("Y") < format.rfind("d"):
                return parse(input_date, dayfirst=False).strftime(format)
            else:
                return parse(input_date, dayfirst=True).strftime(format)
        else:
            return parse(input_date).strftime(format)
    else:
        raise ValueError("[gu.ensure_strftime]: The input date type must be either python datetime, numpy datetime64 or a date formatted string. We got %s" % (type(date)))        
        
def panda2datetime(s):
    """
    This is an extremely fast approach to datetime parsing.
    For large data, the same dates are often repeated. Rather than
    re-parse these, we store all unique dates, parse them, and
    use a lookup to convert all dates.
    
    Ref
    ---
    https://stackoverflow.com/a/29882676/8765762
    https://stackoverflow.com/a/37453925/8765762
    
    """
    dates = {date:pd.to_datetime(date).to_pydatetime() for date in s.unique()}
    dates64 = s.map(dates)
    return dates64
    return np.array([np.datetime64(date) for date in dates64], dtype=object)
            
def BSTCor(time, date):
    """Converts array, time from BST (UTC+1) to GMT (UTC+0) based on the
    daily light saving thresholds for each year which is found from the
    date parameter
    """
    
    import calendar
    
    DST_1 = max(week[-1] for week in calendar.monthcalendar(date.year, 3))
    DST_2 = max(week[-1] for week in calendar.monthcalendar(date.year, 10))

    if date == datetime(date.year, 3, DST_1): #Date lands on daily light saving transition (UTC+0 --> UTC+1)
        print(1)
        time[time >= 2.0] -= 1
    elif date == datetime(date.year, 10, DST_2): #Date lands on daily light saving transition (UTC+1 --> UTC+0)
        print(2)
        time[time < 2.0] -= 1
    elif (date > datetime(date.year, 3, DST_1)) & (date < datetime(date.year, 10, DST_2)): #Date land on UTC+1
        print(3)
        time -= 1
    
    return time
    
def hour2hms(arr):
    """Converts an array containing fractional hours to hours, minutes and seconds in
    integer format"""
    
    #Convert input array
    arr = np.asarray(arr)
    arr = np.atleast_1d(arr)
    
    #Calculate hour
    hour = np.floor(arr).astype(int)
    
    #Calculate minute
    minute = np.floor((arr - hour)*60).astype(int)
    
    #Calculate seconds
    second = np.floor((arr - hour - minute/60.)*3600).astype(int)
    
    return np.vstack((hour, minute, second))
    
def addHourFrac(base_date, fractional_hours):
    """Adds fractional hours to a datetime64 array. This function overcomes the major limitation
    of timedelta64 which can only accept integer values.
    
    Parameters
    ----------
    base_date : np.datetime64
        Must be a single datetime64 object or an array of datetime64 objects of same length
        as fractional_hours.
    fractional_hours : 1D array_like of floats
        An array_like of floats which represent the fractional hours to be added to base_date."""
    
    #Convert fractional_hours to (hour, minute, second)
    HMS = hour2hms(fractional_hours)
    
    #Calculate the timedelta function
    timedelta = HMS[0].astype('timedelta64[h]') + HMS[1].astype('timedelta64[m]') + HMS[2].astype('timedelta64[s]')
    
    #Add timedelta function to base_date
    return base_date + timedelta



from __future__ import absolute_import, division, print_function
import sys
import time, os, warnings
import numpy as np
from datetime import datetime
import matplotlib
import matplotlib.backends
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, FormatStrFormatter
from matplotlib.dates import DateFormatter, MinuteLocator, HourLocator, DayLocator
from mpl_toolkits.axes_grid1 import make_axes_locatable

from .system import isarray, arrlen

# Compatability for python3
if getattr(sys.version_info, 'major') == 3:
	xrange = range

def colorbar(mappable, loc="right", size="5%", pad=0.05):
    """Fixes location of colour bar to the mappable plot
    
    Parameters
    ----------
    mappable : matplotlib object
        The specific matplotlib entity you want a colour bar for. (See example for 
        for details)
    loc : str, optional
        The location you want the colour bar. The possible values are
        
        ["left"|"right"|"bottom"|"top"]
        
        Default is "right"
    size : str, float or int, optional
        Specify the size of the colour bar either as a percentage (str), or the number 
        of pixels width (float or int)
    pad : float, optional
        Specify the amount of white space padding you want between the mappable and the
        colour bar.
        
    Reference
    ---------
    http://joseph-long.com/writing/colorbars/
    
    """
    
    ax = mappable.axes
    fig = ax.figure
    divider = make_axes_locatable(mappable)
    cax = divider.append_axes(loc, size=size, pad=pad)
    
    return fig.colorbar(mappable, cax=cax)
    
def return_intersection(hist_1, hist_2):
    minima = np.minimum(hist_1, hist_2)
    intersection = np.true_divide(np.sum(minima), np.sum(hist_2))
    return intersection

def get_axis_limits(ax, scale=.85, position='upper-right'):
    """Used to position the identifier labels in the corner of images.
    e.g. '(a)' etc. 
    
    An example use with this definition is as follows:
    
        ax1.annotate("("+d[i+1-10]+")", xy=get_axis_limits(ax1))
    
    This will place an annotation in the top right corner of our subplot,
    ax1. Specifically this is used for labelling multiple plots using 
    letters. The function d[i] is a dictionary function which converts
    numbers to letters.
    
    Parameters
    ----------
    ax : plot/subplot function
        The plot or subplot that you want to scale accordingly
    scale : float, optional
        A scale function that determines as a fractional percentage the
        position from the edge of the plot (e.g. scale = 1 is at the
        edge and scale = 0.5 is centred).
    position : string, optional
        Determines the prefered location of interest releative to the
        four corners of the plot. string selections are ('upper-right',
        'upper-left', 'lower-right', 'lower-left')
        
    Returns
    -------
    x : float
        coordinate reference relative to the input plot ax in the x axis
    y: float
        coordinate reference relative to the input plot ax in the y axis
    
    """
    
    if position == 'upper-right':
        try:
            return ax.get_xlim()[1]*0.95*scale, ax.get_ylim()[1]*0.90
        except AttributeError:
            return ax.xlim()[1]*0.95*scale, ax.ylim()[1]*0.90
    elif position == 'upper-left':
        try:
            return ax.get_xlim()[1]*0.95*(1-scale), ax.get_ylim()[1]*0.90
        except AttributeError:
            return ax.xlim()[1]*0.95*(1-scale), ax.ylim()[1]*0.90
    elif position == 'lower-right':
        try:
            return ax.get_xlim()[1]*0.95, ax.get_ylim()[1]*(1-scale)
        except AttributeError:
            return ax.xlim()[1]*0.95, ax.ylim()[1]*(1-scale)
    elif position == 'lower-left':
        try:
            return ax.get_xlim()[1]*0.95, ax.get_ylim()[1]*0.95*(1-scale)
        except AttributeError:
            return ax.xlim()[1]*0.95, ax.ylim()[1]*0.95*(1-scale)
  
def unlink_wrap(dat, lims=[-np.pi, np.pi], thresh = 0.95):
    """
    Iterate over contiguous regions of `dat` (i.e. where it does not
    jump from near one limit to the other).

    This function returns an iterator object that yields slice
    objects, which index the contiguous portions of `dat`.

    This function implicitly assumes that all points in `dat` fall
    within `lims`.

    Ref. https://stackoverflow.com/a/27139390
    
    """
    jump = np.nonzero(np.abs(np.diff(dat)) > ((lims[1] - lims[0]) * thresh))[0]
    lasti = 0
    for ind in jump:
        yield slice(lasti, ind + 1)
        lasti = ind + 1
    yield slice(lasti, len(dat))

def date_ticks(ay, daterange):
    """Fixes ticks for time-series data.
    
    Parameters
    ----------
    ay : Matplotlib axis
        Provide the current axis for your Matplotlib plot that you want to change the axis for. 
        E.g. use ay = plt.gca()
    daterange : tuple, list or ndarray
        The upper and lower date range limits that you want to plot for inside a tuple, list or
        ndarray. E.g. (datetime(2018,1,1,12,0,0), datetime(2018,1,2,0,0,0)). Each element must
        contain either a python datetime or a numpy datetime64 object.    
    """
    
    #Check attributes
    if not isinstance(ay, plt.matplotlib.axes._axes.Axes): raise ValueError("[gu.date_ticks] ay parameter must be a matplotlib subplot. We got %s" % type(ay))
    if isinstance(daterange, (tuple, list, np.ndarray)):
        if len(daterange) != 2: raise ValueError("[gu.date_ticks] daterange parameter must have a length of 2, specifying the date boundaries you are plotting")
        if isinstance(daterange[0], (datetime, np.datetime64)):
            daterange = np.array(daterange).astype(datetime)
        else:
            raise ValueError("[gu.date_ticks] daterange parameter must contain python datetime or numpy datetime64 objects. We got (%s, %s)" % (type(daterange[0]), type(daterange[1])))
    else:
        raise ValueError("[gu.date_ticks] daterange parameter must be either a tuple, list or numpy array. We got %s" % type(daterange))
        
    TimeLength = (daterange[1] - daterange[0]).total_seconds() + 1
    if TimeLength/86400 <= 2:
        """Short Range: ~Single Day"""
        myFmt = DateFormatter('%H:%M')
        ay.xaxis.set_major_formatter(myFmt)
        ay.xaxis.set_major_locator(MinuteLocator(interval=int(np.floor((TimeLength/60)/6))))
    elif TimeLength/86400 <= 7:
        """Medium Range: ~Multiple Days"""
        myFmt = DateFormatter('%Y-%m-%d %H:%M') #Use this when plotting multiple days (e.g. monthly summary)
        ay.xaxis.set_major_formatter(myFmt)
        ay.xaxis.set_major_locator(HourLocator(interval=int(round((TimeLength/3600)/6))))
    else:
        """Long Range: ~Months"""
        myFmt = DateFormatter('%Y-%m-%d') #Use this when plotting multiple days (e.g. monthly summary)
        ay.xaxis.set_major_formatter(myFmt)
        ay.xaxis.set_major_locator(DayLocator(interval=int(round((TimeLength/86400)/6))))
    ay.set_xlabel('Time (UTC) between ' + daterange[0].strftime('%d/%m/%Y %H:%M:%S') + " and " + daterange[1].strftime('%d/%m/%Y %H:%M:%S'))
    
def backend_changer(backend='Qt4Agg', verbose=False):
    """Changes the Matplotlib back-end as certain configurations of code can stop any plots
    being created. E.g. using 'screen' in an interactive terminal requires 'Agg' back-end.
    Running PG_Quickplotter directly from the command line can use the default back-end of 
    'Qt4Agg'
    
    Parameters
    ----------
    backend : str, optional, default = 'Qt4Agg'
        The matplotlib backend name you want to change to. The options available can be found
        by running 
    """
    
    if isinstance(backend, str):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            
            plt.switch_backend(backend)
    else:
        raise ValueError("[_backend_changer] Backend needs to be a string. We got %s. Use _backend_checker to see that available Matplotlib backends for your system." % backend)
    
    if verbose is True: print("[INFO] Matplotlib backend has been changed to %s" % backend)
    
def backend_checker(show_supported=False, show_valid=False):
    """Checks the available back-ends in Matplotlib and then tests each supported back-end
    to see if they can actually be used in the environments configurations"""

    def is_backend_module(fname):
        """Identifies if a filename is a Matplotlib backend module"""
        return fname.startswith('backend_') and fname.endswith('.py')

    def backend_fname_formatter(fname): 
        """Removes the extension of the given filename, then takes away the leading 'backend_'."""
        return os.path.splitext(fname)[0][8:]

    # get the directory where the back-ends live
    backends_dir = os.path.dirname(matplotlib.backends.__file__)

    # filter all files in that directory to identify all files which provide a backend
    backend_fnames = filter(is_backend_module, os.listdir(backends_dir))

    backends = [backend_fname_formatter(fname) for fname in backend_fnames]

    if show_supported is True: print("Supported Backends: \t %s " % backends)

    #Validate Back-ends
    backends_valid = []
    for b in backends:
        try:
            plt.switch_backend(b)
            backends_valid += [b]
        except:
            continue

    if show_valid is True: print("Valid Backends: \t %s" % backends_valid)
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        #Try Back-ends Performance
        backends_available = []
        backends_fps = []
        for b in backends_valid:
            try:
                #plt.ion()                                #Turn on interactive mode
                
                plt.switch_backend(b)                    #Switch back-end to test

                plt.clf()
                tstart = time.time()                       #For time keeping
                x = np.arange(0,2*np.pi,0.01)           #X Array
                line, = plt.plot(x,np.sin(x))
                for i in xrange(1,50):
                    line.set_ydata(np.sin(x+i/10.0))    #Update the data
                    plt.draw()                             #Redraw the canvas

                #plt.ioff()
                
                #If nothing went wrong during FPS check then this is a viable back-end to use
                backends_available.append(b)
                backends_fps.append(50/(time.time()-tstart))
            except:
                pass
    
    #Sort lists
    backends_fps, backends_available = (list(t) for t in zip(*sorted(zip(backends_fps, backends_available), reverse=True)))
    
    print("Available Backends to use with Matplotlib\n-----------------------------------------\n")
    print("Backend         FPS\n---------------------------")
    for backend, fps in zip(backends_available, backends_fps):
        print("%s       \t%.4f" % (backend, fps))

def fixed_aspect_ratio(ax, ratio=1, adjustable=None, xscale='linear', yscale='linear'):
    """
    Set a fixed aspect ratio on Matplotlib plots regardless of axis 
    units. Can handle all scale types (e.g. linear-linear, log-log
    and semilog plots).
    
    Notes
    -----
    This function must be called after all plotting has been completed.
    i.e. just before plt.savefig and time_axis if used.
    
    Reference
    ---------
    https://stackoverflow.com/a/37340384
    https://stackoverflow.com/a/39807336
    https://stackoverflow.com/a/24538164
    """
    
    # Check if minor ticks were enabled
    if len(ax.xaxis.get_minor_ticks()) > 0:
        minor_ticks = True
    else:
        minor_ticks = False

    # Set scales for both axes
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
    
        if xscale == yscale:
            if isinstance(ax, np.ndarray):
                ax = ax.ravel()
                
                for subplot in ax:
                    #Get axis limits
                    xvals, yvals = (subplot.get_xlim(), subplot.get_ylim())
                    
                    #Compute axis range
                    xrange = xvals[1] - xvals[0] if xscale == 'linear' else np.log(xvals[1]) - np.log(xvals[0])
                    yrange = yvals[1] - yvals[0] if yscale == 'linear' else np.log(yvals[1]) - np.log(yvals[0])
                    
                    #Force aspect
                    subplot.set_aspect(ratio*(xrange/yrange), adjustable=adjustable)
            else:
                #Get axis limits
                xvals, yvals = (ax.get_xlim(), ax.get_ylim())
                
                #Compute axis range
                xrange = xvals[1] - xvals[0] if xscale == 'linear' else np.log(xvals[1]) - np.log(xvals[0])
                yrange = yvals[1] - yvals[0] if yscale == 'linear' else np.log(yvals[1]) - np.log(yvals[0])

                #Force aspect
                ax.set_aspect(ratio*(xrange/yrange), adjustable=adjustable)

        # Add support for mixed scale axes in Matplotlib
        elif xscale != yscale:

            # Get the figure size in real coordinates:
            f = plt.gcf()
            fwidth = f.get_figwidth()
            fheight = f.get_figheight()

            # get the axis size and position in relative coordinates
            # this gives a BBox object
            bb = ax.get_position()

            # calculate them into real world coordinates
            axwidth = fwidth * (bb.x1 - bb.x0)
            axheight = fheight * (bb.y1 - bb.y0)

            # if the axis is wider than tall, then it has to be narrower
            if axwidth > axheight:
                # calculate the narrowing relative to the figure
                narrow_by = (axwidth - axheight) / fwidth
                
                # move bounding box edges inwards the same amount to give the correct width
                bb.x0 += narrow_by / 2
                bb.x1 -= narrow_by / 2
                
            # else if the axis is taller than wide, make it vertically smaller
            # works the same as above
            elif axheight > axwidth:
                shrink_by = (axheight - axwidth) / fheight
                bb.y0 += shrink_by / 2
                bb.y1 -= shrink_by / 2
            
            # Set the new position for the plot
            ax.set_position(bb)

        else:
            raise ValueError("Something went really wrong. xscale != yscale and xscale == yscale was found to be both False!")

    # Turn on minor_ticks if previously set
    if minor_ticks:
        ax.minorticks_on()

    return    

def get_aspect(ax=None):
    """Gets aspect ratio of an axes
    
    Ref: https://stackoverflow.com/a/41597178/8765762"""

    if ax is None:
        ax = plt.gca()
    fig = ax.figure

    ll, ur = ax.get_position() * fig.get_size_inches()
    width, height = ur - ll
    axes_ratio = height / width
    aspect = axes_ratio / ax.get_data_ratio()

    return aspect
                        
def fake_colorbar():
    """Creates a fake mappable object used in creating space for a colorbar object. Useful when
    plotting multiple sub-plots with some sub-plots not requiring a colorbar."""
    
    Z = [[0,0],[0,0]]
    levels = np.arange(0,10,1)
    im = plt.contourf(Z, levels)
    
    return im
    
def hide_axis(plt=None, ax=None, x_or_y=None, remove_gridlines=False, remove_labels=False):
    """
    Removes axis of a Matplotlib plot.
    
    Parameters
    ----------
    plt : matplotlib figure
        The figure to modify. N.B. Either plt or ax can be specified,
        not both.
    ax : matplotlib axis
        The axis to modify. N.B. Either plt or ax can be specified,
        not both.
    x_or_y : str
        The x or y coordinates to modify.
    remove_gridlines : bool, optional, default = False
        Specify whether to fully hide the x or y axis. If set to False 
        then just the tick labels are hidden and not the ticks 
        themselves.
    
    References
    ----------
    Stackoverflow : Hiding axis text in matplotlib plots
    """
    
    if plt is None and ax is None: raise ValueError("[gu.hide_axis]: Must specify either plt or ax plots.")
    if plt is not None and ax is not None: raise ValueError("[gu.hide_axis]: Can only specify either plt or ax, not both.")
    if x_or_y is None: raise ValueError("[gu.hide_axis]: x_or_y only takes either 'x' or 'y'.")
    
    if plt is not None: ax = plt.gca()
    
    if x_or_y == 'x':
        if remove_gridlines:
            ax.axes.get_xaxis().set_ticks([])
        elif not remove_gridlines:
            ax.axes.get_xaxis().set_ticklabels([])
        else:
            raise ValueError("remove_gridlines is a boolean parameter and accepts either True or False")
            
        if remove_labels:
            ax.set_xlabel("")
            
    elif x_or_y == 'y':
        if remove_gridlines:
            ax.axes.get_yaxis().set_ticks([])
        elif not remove_gridlines:
            ax.axes.get_yaxis().set_ticklabels([])
        else:
            raise ValueError("remove_gridlines is a boolean parameter and accepts either True or False")
            
        if remove_labels:
            ax.set_ylabel("")
            
    else:
        raise ValueError("x_or_y only takes either 'x' or 'y'.")
        
    return

def time_axis(date_range, plt=None, ax=None, num_ticks=6, auto_format=True, 
        xlabel=None, rotation=None, date_sep="-", time_sep=":"):
    """
    Sets up the x axis using datetime information.
    
    Parameters
    
    """

    # Error check input parameters
    if plt is None and ax is None: 
        raise ValueError("[gu.time_axis]: Must specify either plt or ax plots.")
    if plt is not None and ax is not None: 
        raise ValueError("[gu.time_axis]: Can only specify either plt or ax, not both.")
    if not isinstance(num_ticks, int):
        raise ValueError("[gu.time_axis]: num_ticks must be an integer.")
    if not isarray(date_range):
        raise ValueError("[gu.time_axis]: date_range must be array_like containing two python datetime objects")
    if not isinstance(date_range[0], datetime) or not isinstance(date_range[1], datetime):
        if isinstance(date_range[0], np.datetime64):
            date_range = (date_range[0].astype(datetime),
                    date_range[1].astype(datetime))
        else:
            raise ValueError("[gu.time_axis]: date_range must contain two python datetime objects stating the start and end dates for plotting")
    if len(date_range) != 2:
        raise ValueError("[gu.time_axis]: date_range must have a length of 2. To clarify, date_range must be array_like of length 2 containing python datetime objects which state the start and end times for plotting")
        
    if plt is not None:
        ax = plt.gca()
    
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        
        # If auto has been set for format
        if auto_format:
        
            # Calculate the time length in seconds between datetimes
            time_length = (date_range[1] - date_range[0]).total_seconds() + 1
            
            # Check time_length is bigger than 0
            if time_length == 0:
                raise ValueError("Daterange are too similar. Got interval of 0. Please increase the time between ranges. N.B. time_axis only supports second precision.")
            
            # Specify different tick requirements dependent on the number of seconds between datetimes
            if time_length/86400 <= 2:
                # Short Range: ~Single Day
                myFmt = DateFormatter('%H' + time_sep + '%M')
                ax.xaxis.set_major_formatter(myFmt)
                ax.xaxis.set_major_locator(MinuteLocator(interval=int(np.floor((time_length/60)/num_ticks))))
            elif time_length/86400 <= 7:
                # Medium Range: ~Multiple Days
                myFmt = DateFormatter('%Y' + date_sep + '%m' + date_sep + '%d %H' + time_sep + '%M')
                ax.xaxis.set_major_formatter(myFmt)
                ax.xaxis.set_major_locator(HourLocator(interval=int(round((time_length/3600)/num_ticks))))
            else:
                # Long Range: ~Months
                myFmt = DateFormatter('%Y' + date_sep + '%m' + date_sep + '%d')
                ax.xaxis.set_major_formatter(myFmt)
                ax.xaxis.set_major_locator(DayLocator(interval=int(round((time_length/86400)/num_ticks))))
            
            if xlabel is None:
                ax.set_xlabel('Time (UTC) between ' 
                        + date_range[0].strftime('%Y' + date_sep + '%m' 
                            + date_sep + '%d %H' + time_sep + '%M' 
                            + time_sep + '%S') + " and " 
                        + date_range[1].strftime('%Y' + date_sep + '%m' 
                            + date_sep + '%d %H' + time_sep + '%M' 
                            + time_sep + '%S'))
                            
            elif xlabel is not False:
                ax.set_xlabel(xlabel)
                
        # If format is not set to 'auto'
        else:
            ax.xaxis.set_major_locator(MultipleLocator(num_ticks))
            ax.set_xlabel('Time (UTC)')
    
    # Rotate the ticks as specified by rotation parameter
    if rotation is not None:
        for tick in ax.get_xticklabels():
            tick.set_rotation(rotation)
    
    # Need to draw figure before xtickslabels are populated
    #matplotlib.pyplot.draw()
    return ax.get_xticks(), [item._text for item in ax.get_xticklabels()]

def match_ticks(ax1, ax2, tick_count=5, dec_limit=2, scale='linear'):
    """
    Match the ticks for twin plots nicely
    
    Notes
    -----
    Can't be used in conjunction with fixed_aspect_ratio() for some 
    undetermined reason.
    """
    
    # Vectorise length with numpy (https://stackoverflow.com/a/44587916/8765762)
    arrlen = np.vectorize(len)
    
    ax1.set_yscale(scale)
    ax2.set_yscale(scale)
    
    for i in xrange(10):
        
        # Get y-bounds for ax1 and ax2
        ax1_ybounds = (np.linspace(*ax1.get_ybound(), num=tick_count) % 1).astype(str)
        ax2_ybounds = (np.linspace(*ax2.get_ybound(), num=tick_count) % 1).astype(str)
        
        # Calculate the length of decimal places
        ax1_ybounds_length = arrlen(ax1_ybounds)
        ax2_ybounds_length = arrlen(ax2_ybounds)
        
        # Test is within limits
        if np.all(ax1_ybounds_length - 2 < dec_limit) and np.all(ax2_ybounds_length - 2 < dec_limit):
            
            # Set y-ticks for both plots
            ax1.set_yticks(np.linspace(*ax1.get_ybound(), num=tick_count))
            ax2.set_yticks(np.linspace(*ax2.get_ybound(), num=tick_count))
            
            # Find the maximum number of decimal places
            tick_dec = np.nanmax((ax1_ybounds_length - 2, ax2_ybounds_length - 2)).astype(str)
            
            # Ensure the formatting of the ticks is consistent
            ax1.yaxis.set_major_formatter(FormatStrFormatter('%.' + tick_dec + 'f'))
            ax2.yaxis.set_major_formatter(FormatStrFormatter('%.' + tick_dec + 'f'))
            
            break
            
        else:
            
            tick_count += 1
    
    if i == 9:
        warnings.warn("Could not match up ticks correctly in match_ticks", UserWarning, stacklevel=3)        


from __future__ import absolute_import, division, print_function
import warnings

name = 'gilly_utilities'

__version__ = '0.9.4'
__doc__ = "Gilly Utilities is a module which adds extra support for common data analysis problems. The functions and classes in this module are mainly extensions of the numpy package with some extensions for scipy and pandas."

with warnings.catch_warnings():
	warnings.simplefilter("ignore")

	from .datetime64 import *
	from .nanfunctions import *
	from .math import *
	from .stats import *
	from .meteorology import *
	from .system import *
	from .plotting import *
	from .manipulation import *
	from .geospatial import *
	from .extras import *
	